import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from functools import partial
from jaxmarl.environments.mpe.simple import State, SimpleMPE
from gymnax.environments.spaces import Box
from jaxmarl.environments.mpe.default_params import *


SimpleFacmacMPE3a = lambda: SimpleFacmacMPE(num_good_agents=1, num_adversaries=3, num_landmarks=2,
                            view_radius=1.5, score_function="min")
SimpleFacmacMPE6a = lambda: SimpleFacmacMPE(num_good_agents=2, num_adversaries=6, num_landmarks=4,
                            view_radius=1.5, score_function="min")
SimpleFacmacMPE9a = lambda: SimpleFacmacMPE(num_good_agents=3, num_adversaries=9, num_landmarks=6,
                            view_radius=1.5, score_function="min")

class SimpleFacmacMPE(SimpleMPE):
    def __init__(
        self,
        num_good_agents=1,
        num_adversaries=3,
        num_landmarks=2,
        view_radius=1.5,  # set -1 to deactivate
        score_function="sum"
    ):
        dim_c = 2  # NOTE follows code rather than docs
        action_type = CONTINUOUS_ACT
        view_radius = view_radius if view_radius != -1 else 999999

        num_agents = num_good_agents + num_adversaries
        num_entities = num_agents + num_landmarks

        self.num_good_agents, self.num_adversaries = num_good_agents, num_adversaries

        self.adversaries = ["adversary_{}".format(i) for i in range(num_adversaries)]
        self.good_agents = ["agent_{}".format(i) for i in range(num_good_agents)]
        agents = self.adversaries + self.good_agents

        landmarks = ["landmark {}".format(i) for i in range(num_landmarks)]

        colour = (
            [ADVERSARY_COLOUR] * num_adversaries
            + [AGENT_COLOUR] * num_good_agents
            + [OBS_COLOUR] * num_landmarks
        )

        # Parameters
        rad = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 0.075),
                jnp.full((self.num_good_agents), 0.05),
                jnp.full((num_landmarks), 0.2),
            ]
        )
        accel = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 3.0),
                jnp.full((self.num_good_agents), 4.0),
            ]
        )
        max_speed = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 1.0),
                jnp.full((self.num_good_agents), 1.3),
                jnp.full((num_landmarks), 0.0),
            ]
        )
        collide = jnp.full((num_entities,), True)

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            dim_c=dim_c,
            colour=colour,
            rad=rad,
            accel=accel,
            max_speed=max_speed,
            collide=collide,
        )

        # Overwrite action and observation spaces
        self.observation_spaces = {
            i: Box(-jnp.inf, jnp.inf, (16,)) for i in self.adversaries
        }
        self.action_spaces = {i: Box(0.0, 1.0, (5,)) for i in self.adversaries}

        # Introduce partial observability by limiting the agents' view radii
        self.view_radius = jnp.concatenate(
            [
                jnp.full((num_agents), view_radius),
                jnp.full((num_landmarks), 0.0),
            ]
        )

        self.score_function = score_function

    def rewards(self, state: State) -> Dict[str, float]:
        @partial(jax.vmap, in_axes=(0, None))
        def _collisions(agent_idx: int, other_idx: int):
            return jax.vmap(self.is_collision, in_axes=(None, 0, None))(
                agent_idx,
                other_idx,
                state,
            )

        c = _collisions(
            jnp.arange(self.num_good_agents) + self.num_adversaries,
            jnp.arange(self.num_adversaries),
        )  # [agent, adversary, collison]

        def _good(aidx: int, collisions: chex.Array):
            rew = -10 * jnp.sum(collisions[aidx])

            mr = jnp.sum(self.map_bounds_reward(jnp.abs(state.p_pos[aidx])))
            rew -= mr
            return rew

        ad_rew = 10 * jnp.sum(c)

        rew = {a: ad_rew for a in self.adversaries}
        rew.update(
            {
                a: _good(i + self.num_adversaries, c)
                for i, a in enumerate(self.good_agents)
            }
        )
        return rew

    def _prey_policy(self, key: chex.PRNGKey, state: State, aidx: int):
        action = None
        n = 100  # number of positions sampled
        # sample actions randomly from a target circle
        # length = jnp.sqrt(jnp.random.uniform(0, 1, n))
        # angle = jnp.pi * jnp.random.uniform(0, 2, n)

        key, _key = jax.random.split(key)
        length = jnp.sqrt(jax.random.uniform(_key, (n,), minval=0., maxval=1.))
        key, _key = jax.random.split(key)
        angle = jnp.pi * jnp.sqrt(jax.random.uniform(_key, (n,), minval=0., maxval=2.))
        x = length * jnp.cos(angle)
        y = length * jnp.sin(angle)

        # evaluate score for each position
        # check whether positions are reachable
        # sample a few evenly spaced points on the way and see if they collide with anything
        scores = jnp.zeros(n, dtype=jnp.float32)
        n_iter = 5
        if self.score_function == "sum":
            for i in range(n_iter):
                waypoints_length = (length / float(n_iter)) * (i + 1)
                x_wp = waypoints_length * jnp.cos(angle)
                y_wp = waypoints_length * jnp.sin(angle)
                proj_pos = jnp.vstack((x_wp, y_wp)).transpose() + state.p_pos[aidx]
                delta_pos = state.p_pos[None, :, :] - proj_pos[:, None, :]
                dist = jnp.sqrt(jnp.sum(jnp.square(delta_pos), axis=2))
                dist_min = self.rad + self.rad[aidx]
                scores = jnp.where((dist < dist_min[None]).sum(axis=1), scores, -9999999)
                if i == n_iter - 1:
                    scores += dist[:, :self.num_adversaries].sum(axis=1)
        elif self.score_function == "min":
            proj_pos = jnp.vstack((x, y)).transpose() + state.p_pos[aidx]
            rel_dis = jnp.sqrt(jnp.sum(jnp.square(state.p_pos[aidx] - state.p_pos[:self.num_adversaries])))
            min_dist_adv_idx = jnp.argmin(rel_dis)
            delta_pos = state.p_pos[:self.num_adversaries][None, :, :] - proj_pos[:, None, :]
            dist = jnp.sqrt(jnp.sum(jnp.square(delta_pos), axis=2))
            dist_min = self.rad[:self.num_adversaries] + self.rad[aidx]
            scores = jnp.where((dist < dist_min[None]).sum(axis=1), scores, -9999999)
            scores += dist[:, min_dist_adv_idx]
        else:
            raise Exception("Unknown score function {}".format(self.score_function))
        # move to best position
        best_idx = jnp.argmax(scores)
        chosen_action = jnp.array([x[best_idx], y[best_idx]], dtype=jnp.float32)
        chosen_action = jax.lax.cond(scores[best_idx] < 0, lambda: chosen_action*0.0, lambda: chosen_action)
        return chosen_action

    @partial(jax.jit, static_argnums=[0])
    def step_env(self, key: chex.PRNGKey, state: State, actions: dict):
        u, c = self.set_actions(actions)
        # we throw away num_good_agents now, as num_agents does not differentiate between active and passive agents
        u = u[:-self.num_good_agents]
        for i in range(self.num_good_agents):
            prey_action = self._prey_policy(key, state, u.shape[0]-1+i)
            u = jnp.concatenate([u, prey_action[None]], axis=0)
        if (
            c.shape[1] < self.dim_c
        ):  # This is due to the MPE code carrying around 0s for the communication channels, and due to added prey
            c = jnp.concatenate(
                [c, jnp.zeros((self.num_agents, self.dim_c - c.shape[1]))], axis=1
            )

        key, key_w = jax.random.split(key)
        p_pos, p_vel = self._world_step(key_w, state, u)

        key_c = jax.random.split(key, self.num_agents)
        c = self._apply_comm_action(key_c, c, self.c_noise, self.silent)
        done = jnp.full((self.num_agents), state.step >= self.max_steps)

        state = state.replace(
            p_pos=p_pos,
            p_vel=p_vel,
            c=c,
            done=done,
            step=state.step + 1,
        )

        reward = self.rewards(state)
        obs = self.get_obs(state)

        info = {}

        dones = {a: done[i] for i, a in enumerate(self.agents)}
        dones.update({"__all__": jnp.all(done)})

        return obs, state, reward, dones, info

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        @partial(jax.vmap, in_axes=(0))
        def _common_stats(aidx):
            """Values needed in all observations"""

            landmark_pos = (
                    state.p_pos[self.num_agents:] - state.p_pos[aidx]
            )  # Landmark positions in agent reference frame

            # Zero out unseen agents with other_mask
            other_pos = state.p_pos[: self.num_agents] - state.p_pos[aidx]
            other_vel = state.p_vel[: self.num_agents]

            # use jnp.roll to remove ego agent from other_pos and other_vel arrays
            other_pos = jnp.roll(other_pos, shift=self.num_agents - aidx - 1, axis=0)[
                        : self.num_agents - 1
                        ]
            other_vel = jnp.roll(other_vel, shift=self.num_agents - aidx - 1, axis=0)[
                        : self.num_agents - 1
                        ]

            other_pos = jnp.roll(other_pos, shift=aidx, axis=0)
            other_vel = jnp.roll(other_vel, shift=aidx, axis=0)

            # mask out entities and other agents not in view radius of agent
            landmark_mask = jnp.sqrt(jnp.sum(landmark_pos ** 2)) > self.view_radius[aidx]
            landmark_pos = jnp.where(landmark_mask, 0.0, landmark_pos)

            other_mask = jnp.sqrt(jnp.sum(other_pos ** 2)) > self.view_radius[aidx]
            other_pos = jnp.where(other_mask, 0.0, other_pos)
            other_vel = jnp.where(other_mask, 0.0, other_vel)

            return landmark_pos, other_pos, other_vel

        landmark_pos, other_pos, other_vel = _common_stats(self.agent_range)

        def _good(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                    jnp.zeros_like(other_vel[aidx, -1:].flatten())  # 2 - ensure same dimensionality for all agents!
                ]
            )

        def _adversary(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                    other_vel[aidx, -1:].flatten(),  # 2
                ]
            )

        obs = {a: _adversary(i) for i, a in enumerate(self.adversaries)}
        obs.update(
            {a: _good(i + self.num_adversaries) for i, a in enumerate(self.good_agents)}
        )
        return obs

    def get_world_state(self, state: State):
        @partial(jax.vmap, in_axes=(0))
        def _common_stats(aidx):
            """Values needed in all observations"""

            landmark_pos = (
                    state.p_pos[self.num_agents:] - state.p_pos[aidx]
            )  # Landmark positions in agent reference frame

            # Zero out unseen agents with other_mask
            other_pos = state.p_pos[: self.num_agents] - state.p_pos[aidx]
            other_vel = state.p_vel[: self.num_agents]

            # use jnp.roll to remove ego agent from other_pos and other_vel arrays
            other_pos = jnp.roll(other_pos, shift=self.num_agents - aidx - 1, axis=0)[
                        : self.num_agents - 1
                        ]
            other_vel = jnp.roll(other_vel, shift=self.num_agents - aidx - 1, axis=0)[
                        : self.num_agents - 1
                        ]

            other_pos = jnp.roll(other_pos, shift=aidx, axis=0)
            other_vel = jnp.roll(other_vel, shift=aidx, axis=0)

            return landmark_pos, other_pos, other_vel

        landmark_pos, other_pos, other_vel = _common_stats(self.agent_range)

        def _good(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                ]
            )

        def _adversary(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                    other_vel[aidx, -1:].flatten(),  # 2
                ]
            )

        obs = {a: _adversary(i) for i, a in enumerate(self.adversaries)}
        obs.update(
            {a: _good(i + self.num_adversaries) for i, a in enumerate(self.good_agents)}
        )
        return obs

if __name__ == "__main__":
    env = SimpleFacmacMPE(0)
    vec_step_env = jax.jit(env.step_env)
    jax.jit(env.step_env)
    import jaxmarl
    env = jaxmarl.make("MPE_simple_pred_prey_v1")
    get_obs = jax.jit(env.get_obs)

    num_envs = 128
    rng = jax.random.PRNGKey(30)
    rng, _rng = jax.random.split(rng)
    reset_rng = jax.random.split(_rng, num_envs)
    obsv, env_state = jax.vmap(env.reset)(reset_rng)

    rng, _rng = jax.random.split(rng)
    rng_step = jax.random.split(_rng, num_envs)
    env_act = jnp.zeros((num_envs, 1))
    obsv, env_state, reward, done, info = jax.vmap(env.step)(rng_step, env_state, env_act)
    pass



import chex
import jax.numpy as jnp
import jax.tree_util


def tile_over_axis(tensor, axis, size):
    tensor = jnp.expand_dims(tensor, axis)
    tensor = jnp.repeat(tensor, size, axis=axis)
    return tensor


def type_to_dtype(t):
    if t == int:
        return jnp.int32
    elif t == float:
        return jnp.float32
    elif t == bool:
        return jnp.bool_
    else:
        raise ValueError(f"Unsupported type: {t}")


def scalar_to_jax(scalar):
    chex.assert_scalar(scalar)
    return jnp.asarray(scalar, dtype=type_to_dtype(type(scalar)))


def assert_tree_is_batch_of_tree(tree_batch, tree):
    def cmp(tensor_batch, tensor):
        return tensor_batch.shape[1:] == tensor.shape

    def error(tensor_batch, tensor):
        return f"{tensor_batch.shape[1:]} != {tensor.shape}"

    chex.assert_trees_all_equal_comparator(cmp, error, tree_batch, tree)


def set_pytree_batch_item(tree_batch, index, tree):
    assert_tree_is_batch_of_tree(tree_batch, tree)

    return jax.tree_util.tree_map(
        lambda tb, t: tb.at[index].set(t),
        tree_batch, tree,
    )


def get_pytree_batch_item(tree_batch, index):
    return jax.tree_util.tree_map(lambda tb: tb[index], tree_batch)


def get_pytree_axis_dim(tree, axis):
    leaves = jax.tree_util.tree_leaves(tree)
    assert len(leaves) > 0
    axis_dim = leaves[0].shape[axis]
    assert all(leave.shape[axis] == axis_dim for leave in leaves)
    return axis_dim


def concatenate_trees(*trees, axis=0):
    return jax.tree_util.tree_map(lambda *t: jnp.concatenate(t, axis=axis), *trees)


def stack_trees(*trees, axis=0):
    return jax.tree_util.tree_map(lambda *t: jnp.stack(t, axis=axis), *trees)



""" Wrappers for use with jaxmarl baselines. """
import jax
import jax.numpy as jnp
import chex
import numpy as np
from flax import struct
from functools import partial

# from gymnax.environments import environment, spaces
from typing import Optional, Tuple, Union
from jaxmarl.environments.multi_agent_env import MultiAgentEnv, State


class JaxMARLWrapper(object):
    """Base class for all jaxmarl wrappers."""

    def __init__(self, env: MultiAgentEnv):
        self._env = env

    def __getattr__(self, name: str):
        return getattr(self._env, name)

    # def _batchify(self, x: dict):
    #     x = jnp.stack([x[a] for a in self._env.agents])
    #     return x.reshape((self._env.num_agents, -1))

    def _batchify_floats(self, x: dict):
        return jnp.stack([x[a] for a in self._env.agents])


@struct.dataclass
class LogEnvState:
    env_state: State
    episode_returns: float
    episode_lengths: int
    returned_episode_returns: float
    returned_episode_lengths: int


class LogWrapper(JaxMARLWrapper):
    """Log the episode returns and lengths.
    NOTE for now for envs where agents terminate at the same time.
    """

    def __init__(self, env: MultiAgentEnv, replace_info: bool = False):
        super().__init__(env)
        self.replace_info = replace_info

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        obs, env_state = self._env.reset(key)
        state = LogEnvState(
            env_state,
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
        )
        return obs, state

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: LogEnvState,
        action: Union[int, float],
    ) -> Tuple[chex.Array, LogEnvState, float, bool, dict]:
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action
        )
        ep_done = done["__all__"]
        new_episode_return = state.episode_returns + self._batchify_floats(reward)
        new_episode_length = state.episode_lengths + 1
        state = LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - ep_done),
            episode_lengths=new_episode_length * (1 - ep_done),
            returned_episode_returns=state.returned_episode_returns * (1 - ep_done)
            + new_episode_return * ep_done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - ep_done)
            + new_episode_length * ep_done,
        )
        if self.replace_info:
            info = {}
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_episode"] = jnp.full((self._env.num_agents,), ep_done)
        return obs, state, reward, done, info

class MPELogWrapper(LogWrapper):
    """ Times reward signal by number of agents within the environment,
    to match the on-policy codebase. """
    
    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: LogEnvState,
        action: Union[int, float],
    ) -> Tuple[chex.Array, LogEnvState, float, bool, dict]:
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action
        )
        rewardlog = jax.tree_map(lambda x: x*self._env.num_agents, reward)  # As per on-policy codebase
        ep_done = done["__all__"]
        new_episode_return = state.episode_returns + self._batchify_floats(rewardlog)
        new_episode_length = state.episode_lengths + 1
        state = LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - ep_done),
            episode_lengths=new_episode_length * (1 - ep_done),
            returned_episode_returns=state.returned_episode_returns * (1 - ep_done)
            + new_episode_return * ep_done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - ep_done)
            + new_episode_length * ep_done,
        )
        if self.replace_info:
            info = {}
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_episode"] = jnp.full((self._env.num_agents,), ep_done)
        return obs, state, reward, done, info

@struct.dataclass
class SMAXLogEnvState:
    env_state: State
    episode_returns: float
    episode_lengths: int
    won_episode: int
    returned_episode_returns: float
    returned_episode_lengths: int
    returned_won_episode: int


class SMAXLogWrapper(JaxMARLWrapper):
    def __init__(self, env: MultiAgentEnv, replace_info: bool = False):
        super().__init__(env)
        self.replace_info = replace_info

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        obs, env_state = self._env.reset(key)
        state = SMAXLogEnvState(
            env_state,
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
        )
        return obs, state

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: SMAXLogEnvState,
        action: Union[int, float],
    ) -> Tuple[chex.Array, LogEnvState, float, bool, dict]:
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action
        )
        ep_done = done["__all__"]
        batch_reward = self._batchify_floats(reward)
        new_episode_return = state.episode_returns + self._batchify_floats(reward)
        new_episode_length = state.episode_lengths + 1
        new_won_episode = (batch_reward >= 1.0).astype(jnp.float32)
        state = SMAXLogEnvState(
            env_state=env_state,
            won_episode=new_won_episode * (1 - ep_done),
            episode_returns=new_episode_return * (1 - ep_done),
            episode_lengths=new_episode_length * (1 - ep_done),
            returned_episode_returns=state.returned_episode_returns * (1 - ep_done)
            + new_episode_return * ep_done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - ep_done)
            + new_episode_length * ep_done,
            returned_won_episode=state.returned_won_episode * (1 - ep_done)
            + new_won_episode * ep_done,
        )
        if self.replace_info:
            info = {}
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_won_episode"] = state.returned_won_episode
        info["returned_episode"] = jnp.full((self._env.num_agents,), ep_done)
        return obs, state, reward, done, info



from typing import Tuple, Union, Sequence
from collections import OrderedDict
import chex
import jax
import jax.numpy as jnp

class Space(object):
    """
    Minimal jittable class for abstract jaxmarl space.
    """

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        raise NotImplementedError

    def contains(self, x: jnp.int_) -> bool:
        raise NotImplementedError

class Discrete(Space):
	"""
	Minimal jittable class for discrete gymnax spaces.
	TODO: For now this is a 1d space. Make composable for multi-discrete.
	"""

	def __init__(self, num_categories: int, dtype=jnp.int32):
		assert num_categories >= 0
		self.n = num_categories
		self.shape = ()
		self.dtype = dtype

	def sample(self, rng: chex.PRNGKey) -> chex.Array:
		"""Sample random action uniformly from set of categorical choices."""
		return jax.random.randint(
			rng, shape=self.shape, minval=0, maxval=self.n
		).astype(self.dtype)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether specific object is within space."""
		# type_cond = isinstance(x, self.dtype)
		# shape_cond = (x.shape == self.shape)
		range_cond = jnp.logical_and(x >= 0, x < self.n)
		return range_cond


class MultiDiscrete(Space):
    """
    Minimal jittable class for multi-discrete gymnax spaces.
    """

    def __init__(self, num_categories: Sequence[int]):
        """Num categories is the number of cat actions for each dim, [2,2,2]=2 actions x 3 dim"""
        self.num_categories = jnp.array(num_categories)
        self.shape = (len(num_categories),)
        self.dtype = jnp.int_

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        """Sample random action uniformly from set of categorical choices."""
        return jax.random.randint(
            rng, 
            shape=self.shape, 
            minval=0, 
            maxval=self.num_categories,
            dtype=self.dtype
        )

    def contains(self, x: jnp.int_) -> bool:
        """Check whether specific object is within space."""
        range_cond = jnp.logical_and(x >= 0, x < self.num_categories)
        return jnp.all(range_cond)


class Box(Space):
	"""
	Minimal jittable class for array-shaped gymnax spaces.
	TODO: Add unboundedness - sampling from other distributions, etc.
	"""
	def __init__(
		self,
		low: float,
		high: float,
		shape: Tuple[int],
		dtype: jnp.dtype = jnp.float32,
	):
		self.low = low
		self.high = high
		self.shape = shape
		self.dtype = dtype

	def sample(self, rng: chex.PRNGKey) -> chex.Array:
		"""Sample random action uniformly from 1D continuous range."""
		return jax.random.uniform(
			rng, shape=self.shape, minval=self.low, maxval=self.high
		).astype(self.dtype)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether specific object is within space."""
		# type_cond = isinstance(x, self.dtype)
		# shape_cond = (x.shape == self.shape)
		range_cond = jnp.logical_and(
			jnp.all(x >= self.low), jnp.all(x <= self.high)
		)
		return range_cond


class Dict(Space):
	"""Minimal jittable class for dictionary of simpler jittable spaces."""
	def __init__(self, spaces: dict):
		self.spaces = spaces
		self.num_spaces = len(spaces)

	def sample(self, rng: chex.PRNGKey) -> dict:
		"""Sample random action from all subspaces."""
		key_split = jax.random.split(rng, self.num_spaces)
		return OrderedDict(
			[
				(k, self.spaces[k].sample(key_split[i]))
				for i, k in enumerate(self.spaces)
			]
		)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether dimensions of object are within subspace."""
		# type_cond = isinstance(x, dict)
		# num_space_cond = len(x) != len(self.spaces)
		# Check for each space individually
		out_of_space = 0
		for k, space in self.spaces.items():
			out_of_space += 1 - space.contains(getattr(x, k))
		return out_of_space == 0


class Tuple(Space):
	"""Minimal jittable class for tuple (product) of jittable spaces."""
	def __init__(self, spaces: Union[tuple, list]):
		self.spaces = spaces
		self.num_spaces = len(spaces)

	def sample(self, rng: chex.PRNGKey) -> Tuple[chex.Array]:
		"""Sample random action from all subspaces."""
		key_split = jax.random.split(rng, self.num_spaces)
		return tuple(
			[
				space.sample(key_split[i])
				for i, space in enumerate(self.spaces)
			]
		)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether dimensions of object are within subspace."""
		# type_cond = isinstance(x, tuple)
		# num_space_cond = len(x) != len(self.spaces)
		# Check for each space individually
		out_of_space = 0
		for space in self.spaces:
			out_of_space += 1 - space.contains(x)
		return out_of_space == 0



import jax.numpy as jnp
import jax
from jaxmarl.environments.multi_agent_env import MultiAgentEnv
from jaxmarl.environments.spaces import Box, Discrete
from jaxmarl.environments.smax.distributions import (
    SurroundAndReflectPositionDistribution,
    UniformUnitTypeDistribution,
)
import chex
from typing import Tuple, Dict, Optional
from flax.struct import dataclass
from enum import IntEnum
from functools import partial
import io


@dataclass
class State:
    unit_positions: chex.Array
    unit_alive: chex.Array
    unit_teams: chex.Array
    unit_health: chex.Array
    unit_types: chex.Array
    unit_weapon_cooldowns: chex.Array
    prev_actions: chex.Array
    time: int
    terminal: bool


@dataclass
class WorldDelta:
    """Encapsulates the effect of an agent's action"""

    pos: chex.Array
    attacked_idx: int
    cooldown_diff: float
    health_diff: float


@dataclass
class Scenario:
    unit_types: chex.Array
    num_allies: int
    num_enemies: int
    smacv2_position_generation: bool
    smacv2_unit_type_generation: bool


MAP_NAME_TO_SCENARIO = {
    # name: (unit_types, n_allies, n_enemies, SMACv2 position generation, SMACv2 unit generation)
    "3m": Scenario(jnp.zeros((6,), dtype=jnp.uint8), 3, 3, False, False),
    "2s3z": Scenario(
        jnp.array([2, 2, 3, 3, 3] * 2, dtype=jnp.uint8), 5, 5, False, False
    ),
    "25m": Scenario(jnp.zeros((50,), dtype=jnp.uint8), 25, 25, False, False),
    "3s5z": Scenario(
        jnp.array(
            [
                2,
                2,
                2,
                3,
                3,
                3,
                3,
                3,
            ]
            * 2,
            dtype=jnp.uint8,
        ),
        8,
        8,
        False,
        False,
    ),
    "8m": Scenario(jnp.zeros((16,), dtype=jnp.uint8), 8, 8, False, False),
    "5m_vs_6m": Scenario(jnp.zeros((11,), dtype=jnp.uint8), 5, 6, False, False),
    "10m_vs_11m": Scenario(jnp.zeros((21,), dtype=jnp.uint8), 10, 11, False, False),
    "27m_vs_30m": Scenario(jnp.zeros((57,), dtype=jnp.uint8), 27, 30, False, False),
    "3s5z_vs_3s6z": Scenario(
        jnp.concatenate(
            [
                jnp.array([2, 2, 2, 3, 3, 3, 3, 3], dtype=jnp.uint8),
                jnp.array([2, 2, 2, 3, 3, 3, 3, 3, 3], dtype=jnp.uint8),
            ]
        ),
        8,
        9,
        False,
        False,
    ),
    "3s_vs_5z": Scenario(
        jnp.array([2, 2, 2, 3, 3, 3, 3, 3], dtype=jnp.uint8), 3, 5, False, False
    ),
    "6h_vs_8z": Scenario(
        jnp.array([5, 5, 5, 5, 5, 5, 3, 3, 3, 3, 3, 3, 3, 3], dtype=jnp.uint8),
        6,
        8,
        False,
        False,
    ),
    "smacv2_5_units": Scenario(jnp.zeros((10,), dtype=jnp.uint8), 5, 5, True, True),
    "smacv2_10_units": Scenario(jnp.zeros((20,), dtype=jnp.uint8), 10, 10, True, True),
    "smacv2_20_units": Scenario(jnp.zeros((40,), dtype=jnp.uint8), 20, 20, True, True),
}


def map_name_to_scenario(map_name):
    """maps from smac map names to a scenario array"""
    return MAP_NAME_TO_SCENARIO[map_name]


def register_scenario(map_name, scenario):
    MAP_NAME_TO_SCENARIO[map_name] = scenario


class SMAX(MultiAgentEnv):
    def __init__(
        self,
        num_allies=5,
        num_enemies=5,
        map_width=32,
        map_height=32,
        world_steps_per_env_step=8,
        time_per_step=1.0 / 16,
        scenario=None,
        unit_type_names=[
            "marine",
            "marauder",
            "stalker",
            "zealot",
            "zergling",
            "hydralisk",
        ],
        unit_type_shorthands=["m", "M", "s", "Z", "z", "h"],
        unit_type_velocities=jnp.array([3.15, 2.25, 4.13, 3.15, 4.13, 3.15]),
        unit_type_attacks=jnp.array([9.0, 10.0, 13.0, 8.0, 5.0, 12.0]),
        unit_type_attack_ranges=jnp.array([5.0, 6.0, 6.0, 2.0, 2.0, 5.0]),
        unit_type_sight_ranges=jnp.array([9.0, 10.0, 10.0, 9.0, 8.0, 9.0]),
        unit_type_radiuses=jnp.array([0.375, 0.5625, 0.625, 0.5, 0.375, 0.625]),
        unit_type_health=jnp.array([45.0, 125.0, 160, 150, 35, 80]),
        unit_type_weapon_cooldowns=jnp.array([0.61, 1.07, 1.87, 0.86, 0.5, 0.59]),
        use_self_play_reward=False,
        see_enemy_actions=True,
        won_battle_bonus=1.0,
        walls_cause_death=True,
        max_steps=100,
        smacv2_position_generation=False,
        smacv2_unit_type_generation=False,
    ) -> None:
        self.num_allies = num_allies if scenario is None else scenario.num_allies
        self.num_enemies = num_enemies if scenario is None else scenario.num_enemies
        self.num_agents = self.num_allies + self.num_enemies
        self.walls_cause_death = walls_cause_death
        self.unit_type_names = unit_type_names
        self.unit_type_shorthands = unit_type_shorthands
        self.num_movement_actions = 5  # 5 cardinal directions + stop
        self.world_steps_per_env_step = world_steps_per_env_step
        self.map_width = map_width
        self.map_height = map_height
        self.scenario = scenario if scenario is None else scenario.unit_types
        self.use_self_play_reward = use_self_play_reward
        self.time_per_step = time_per_step
        self.unit_type_velocities = unit_type_velocities
        self.unit_type_weapon_cooldowns = unit_type_weapon_cooldowns
        self.unit_type_attacks = unit_type_attacks
        self.unit_type_attack_ranges = unit_type_attack_ranges
        self.unit_type_sight_ranges = unit_type_sight_ranges
        self.unit_type_radiuses = unit_type_radiuses
        self.unit_type_health = unit_type_health
        self.unit_type_bits = len(self.unit_type_names)
        self.max_steps = max_steps
        self.won_battle_bonus = won_battle_bonus
        self.see_enemy_actions = see_enemy_actions
        self.smacv2_unit_type_generation = (
            smacv2_unit_type_generation
            if scenario is None
            else scenario.smacv2_unit_type_generation
        )
        self.smacv2_position_generation = (
            smacv2_position_generation
            if scenario is None
            else scenario.smacv2_position_generation
        )
        self.position_generator = SurroundAndReflectPositionDistribution(
            self.num_allies, self.num_enemies, self.map_width, self.map_height
        )
        self.unit_type_generator = UniformUnitTypeDistribution(
            self.num_allies,
            self.num_enemies,
            self.map_width,
            self.map_height,
            len(self.unit_type_names),
        )
        self.agents = [f"ally_{i}" for i in range(self.num_allies)] + [
            f"enemy_{i}" for i in range(self.num_enemies)
        ]
        self.agent_ids = {agent: i for i, agent in enumerate(self.agents)}
        self.teams = jnp.zeros((self.num_agents,), dtype=jnp.uint8)
        self.teams = self.teams.at[self.num_allies :].set(1)
        self.own_features = ["health", "position_x", "position_y", "weapon_cooldown"]
        self.own_features += [f"unit_type_bit_{i}" for i in range(self.unit_type_bits)]
        self.unit_features = [
            "health",
            "position_x",
            "position_y",
            "last_action",
            "weapon_cooldown",
        ]
        self.unit_features += [
            f"unit_type_bits_{i}" for i in range(self.unit_type_bits)
        ]
        self.obs_size = (
            len(self.unit_features) * (self.num_allies - 1)
            + len(self.unit_features) * self.num_enemies
            + len(self.own_features)
        )
        self.state_size = (len(self.own_features) + 2) * self.num_agents
        self.observation_spaces = {
            i: Box(low=0.0, high=1.0, shape=(self.obs_size,)) for i in self.agents
        }
        self.num_ally_actions = self.num_enemies + self.num_movement_actions
        self.num_enemy_actions = self.num_allies + self.num_movement_actions
        self.action_spaces = {
            agent: Discrete(
                num_categories=self.num_ally_actions
                if i < self.num_allies
                else self.num_enemy_actions
            )
            for i, agent in enumerate(self.agents)
        }

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[Dict[str, chex.Array], State]:
        """Environment-specific reset."""
        key, team_0_key, team_1_key = jax.random.split(key, num=3)
        team_0_start = jnp.stack([jnp.array([8.0, 16.0])] * self.num_allies)
        team_0_start_noise = jax.random.uniform(
            team_0_key, shape=(self.num_allies, 2), minval=-2, maxval=2
        )
        team_0_start = team_0_start + team_0_start_noise
        team_1_start = jnp.stack([jnp.array([24.0, 16.0])] * self.num_enemies)
        team_1_start_noise = jax.random.uniform(
            team_1_key, shape=(self.num_enemies, 2), minval=-2, maxval=2
        )
        team_1_start = team_1_start + team_1_start_noise
        unit_positions = jnp.concatenate([team_0_start, team_1_start])
        key, pos_key = jax.random.split(key)
        generated_unit_positions = self.position_generator.generate(pos_key)
        unit_positions = jax.lax.select(
            self.smacv2_position_generation, generated_unit_positions, unit_positions
        )
        unit_teams = jnp.zeros((self.num_agents,))
        unit_teams = unit_teams.at[self.num_allies :].set(1)
        unit_weapon_cooldowns = jnp.zeros((self.num_agents,))
        # default behaviour spawn all marines
        unit_types = (
            jnp.zeros((self.num_agents,), dtype=jnp.uint8)
            if self.scenario is None
            else self.scenario
        )
        key, unit_type_key = jax.random.split(key)
        generated_unit_types = self.unit_type_generator.generate(unit_type_key)
        unit_types = jax.lax.select(
            self.smacv2_unit_type_generation, generated_unit_types, unit_types
        )
        unit_health = self.unit_type_health[unit_types]
        state = State(
            unit_positions=unit_positions,
            unit_alive=jnp.ones((self.num_agents,), dtype=jnp.bool_),
            unit_teams=unit_teams,
            unit_health=unit_health,
            unit_types=unit_types,
            prev_actions=jnp.zeros((self.num_agents,), dtype=jnp.int32),
            time=0,
            terminal=False,
            unit_weapon_cooldowns=unit_weapon_cooldowns,
        )
        state = self._push_units_away(state)
        obs = self.get_obs(state)
        world_state = self.get_world_state(state)
        obs["world_state"] = jax.lax.stop_gradient(world_state)
        return obs, state

    @partial(jax.jit, static_argnums=(0,))
    def step_env(
        self,
        key: chex.PRNGKey,
        state: State,
        actions: Dict[str, chex.Array],
    ) -> Tuple[Dict[str, chex.Array], State, Dict[str, float], Dict[str, bool], Dict]:
        """Environment-specific step transition."""
        actions = jnp.array([actions[i] for i in self.agents])
        health_before = jnp.copy(state.unit_health)

        def world_step_fn(carry, _):
            state, step_key = carry
            step_key, world_step_key = jax.random.split(step_key)
            state = partial(self._world_step, actions=actions)(
                key=world_step_key, state=state
            )
            state = self._kill_agents_touching_walls(state)
            state = self._update_dead_agents(state)
            state = self._push_units_away(state)
            return (state, step_key), None

        (state, _), _ = jax.lax.scan(
            world_step_fn,
            init=(state, key),
            xs=None,
            length=self.world_steps_per_env_step,
        )
        health_after = state.unit_health
        state = state.replace(
            terminal=self.is_terminal(state), prev_actions=actions, time=state.time + 1
        )
        obs = self.get_obs(state)
        dones = {
            agent: ~state.unit_alive[self.agent_ids[agent]] for agent in self.agents
        }
        rewards = self.compute_reward(state, health_before, health_after)
        dones["__all__"] = state.terminal
        world_state = self.get_world_state(state)
        infos = {}
        obs["world_state"] = jax.lax.stop_gradient(world_state)
        return (
            jax.lax.stop_gradient(obs),
            jax.lax.stop_gradient(state),
            rewards,
            dones,
            infos,
        )

    @partial(jax.jit, static_argnums=(0,))
    def compute_reward(self, state, health_before, health_after):
        @partial(jax.jit, static_argnums=(0,))
        def compute_team_reward(team_idx):
            # compute how much the enemy team health has decreased
            other_team_idx = jnp.logical_not(team_idx).astype(jnp.uint32)
            other_team_start_idx = jnp.array([0, self.num_allies])[other_team_idx]
            team_start_idx = jnp.array([0, self.num_allies])[team_idx]

            team_size = self.num_allies if team_idx == 0 else self.num_enemies

            enemy_team_size = self.num_enemies if team_idx == 0 else self.num_allies

            enemy_health_decrease = jnp.sum(
                jax.lax.dynamic_slice_in_dim(
                    (health_after - health_before)
                    / self.unit_type_health[state.unit_types],
                    other_team_start_idx,
                    enemy_team_size,
                )
            )
            enemy_health_decrease_reward = (
                jnp.abs(enemy_health_decrease) / enemy_team_size
            )
            enemy_health_decrease_reward = jax.lax.select(
                self.use_self_play_reward, 0.0, enemy_health_decrease_reward
            )
            won_battle = jnp.all(
                jnp.logical_not(
                    jax.lax.dynamic_slice_in_dim(
                        state.unit_alive, other_team_start_idx, enemy_team_size
                    )
                )
            )
            lost_battle = jnp.all(
                jnp.logical_not(
                    jax.lax.dynamic_slice_in_dim(
                        state.unit_alive, team_start_idx, team_size
                    )
                )
            )
            # have a lost battle bonus in addition to the won bonus in
            # order to make the game zero-sum in self-play and therefore prevent any
            # collaboration.
            lost_battle_bonus = jax.lax.cond(
                lost_battle & self.use_self_play_reward & ~won_battle,
                lambda: -self.won_battle_bonus,
                lambda: 0.0,
            )
            # only award the won_battle_bonus when all the enemy is dead
            # AND there is at least one ally alive. Otherwise it's a draw.
            # This can't happen in SC2 because actions happen in a random order,
            # but I'd rather VMAP over events where possible, which means we
            # can get draws.
            won_battle_bonus = jax.lax.cond(
                won_battle & ~lost_battle, lambda: self.won_battle_bonus, lambda: 0.0
            )
            return enemy_health_decrease_reward + won_battle_bonus + lost_battle_bonus

        # agents still get reward when they are dead to allow for noble sacrifice
        team_rewards = [compute_team_reward(i) for i in range(2)]
        return {
            agent: team_rewards[int(self.agent_ids[agent] >= self.num_allies)]
            for agent in self.agents
        }

    @partial(jax.jit, static_argnums=(0,))
    def is_terminal(self, state):
        all_dead = jnp.all(jnp.logical_not(state.unit_alive[: self.num_allies]))
        all_enemy_dead = jnp.all(jnp.logical_not(state.unit_alive[self.num_allies :]))
        over_time_limit = state.time >= self.max_steps
        return all_dead | all_enemy_dead | over_time_limit

    def _update_dead_agents(
        self,
        state: State,
    ):
        unit_alive = state.unit_health > 0
        return state.replace(unit_alive=unit_alive)

    def _kill_agents_touching_walls(self, state: State):
        units_touching_walls = jnp.logical_or(
            jnp.any(state.unit_positions <= 0.0, axis=-1),
            jnp.any(
                state.unit_positions >= jnp.array([self.map_width, self.map_height]),
                axis=-1,
            ),
        )
        unit_health = jnp.where(units_touching_walls, 0.0, state.unit_health)
        unit_health = jax.lax.select(
            self.walls_cause_death, unit_health, state.unit_health
        )
        return state.replace(unit_health=unit_health)

    def _push_units_away(self, state: State, firmness: float = 1.0):
        delta_matrix = state.unit_positions[:, None] - state.unit_positions[None, :]
        dist_matrix = (
            jnp.linalg.norm(delta_matrix, axis=-1)
            + jnp.identity(self.num_agents)
            + 1e-6
        )
        radius_matrix = (
            self.unit_type_radiuses[state.unit_types][:, None]
            + self.unit_type_radiuses[state.unit_types][None, :]
        )
        overlap_term = jax.nn.relu(radius_matrix / dist_matrix - 1.0)
        unit_positions = (
            state.unit_positions
            + firmness * jnp.sum(delta_matrix * overlap_term[:, :, None], axis=1) / 2
        )
        return state.replace(unit_positions=unit_positions)

    @partial(jax.jit, static_argnums=(0,))
    def _world_step(
        self,
        key: chex.PRNGKey,
        state: State,
        actions: Dict[str, chex.Array],
    ) -> Tuple[Dict[str, chex.Array], State, Dict[str, float], Dict[str, bool], Dict]:
        def update_position(idx, action):
            # Compute the movements slightly strangely.
            # The velocities below are for diagonal directions
            # because these are easier to encode as actions than the four
            # diagonal directions. Then rotate the velocity 45
            # degrees anticlockwise to compute the movement.
            pos = state.unit_positions[idx]
            vec = jax.lax.cond(
                # action is an attack action OR stop (action 4)
                action >= self.num_movement_actions - 1,
                lambda: jnp.zeros((2,)),
                lambda: jnp.array(
                    [
                        (-1) ** (action // 2) * (1.0 / jnp.sqrt(2)),
                        (-1) ** (action // 2 + action % 2) * (1.0 / jnp.sqrt(2)),
                    ]
                ),
            )
            rotation = jnp.array(
                [
                    [1.0 / jnp.sqrt(2), -1.0 / jnp.sqrt(2)],
                    [1.0 / jnp.sqrt(2), 1.0 / jnp.sqrt(2)],
                ]
            )
            vec = rotation @ vec
            new_pos = (
                pos
                + vec
                * self.unit_type_velocities[state.unit_types[idx]]
                * self.time_per_step
            )
            # avoid going out of bounds
            new_pos = jnp.maximum(
                jnp.minimum(new_pos, jnp.array([self.map_width, self.map_height])),
                jnp.zeros((2,)),
            )
            return WorldDelta(new_pos, idx, -self.time_per_step, 0.0)

        def update_agent_health(idx, action, key):
            # for team 1, their attack actions are labelled in
            # reverse order because that is the order they are
            # observed in
            attacked_idx = jax.lax.cond(
                idx < self.num_allies,
                lambda: action + self.num_allies - self.num_movement_actions,
                lambda: self.num_allies - 1 - (action - self.num_movement_actions),
            )
            attack_valid = (
                (
                    jnp.linalg.norm(
                        state.unit_positions[idx] - state.unit_positions[attacked_idx]
                    )
                    < self.unit_type_attack_ranges[state.unit_types[idx]]
                )
                & state.unit_alive[idx]
                & state.unit_alive[attacked_idx]
            )
            attack_valid = attack_valid & (state.unit_weapon_cooldowns[idx] <= 0.0)
            health_diff = jax.lax.select(
                attack_valid,
                -self.unit_type_attacks[state.unit_types[idx]],
                0.0,
            )
            # design choice based on the pysc2 randomness details.
            # See https://github.com/deepmind/pysc2/blob/master/docs/environment.md#determinism-and-randomness

            cooldown_deviation = jax.random.uniform(
                key, minval=-self.time_per_step, maxval=2 * self.time_per_step
            )
            cooldown = (
                self.unit_type_weapon_cooldowns[state.unit_types[idx]]
                + cooldown_deviation
            )
            cooldown_diff = jax.lax.select(
                attack_valid,
                cooldown - state.unit_weapon_cooldowns[idx],
                -self.time_per_step,
            )
            return WorldDelta(
                state.unit_positions[idx],
                attacked_idx,
                cooldown_diff,
                health_diff,
            )

        def perform_agent_action(idx, action, key):
            return jax.lax.cond(
                actions[idx] > self.num_movement_actions - 1,
                lambda: update_agent_health(idx, action, key),
                lambda: update_position(idx, action),
            )

        keys = jax.random.split(key, num=self.num_agents)
        deltas = jax.vmap(perform_agent_action)(
            jnp.arange(self.num_agents), actions, keys
        )

        def update_health(carry: chex.Array, delta: WorldDelta):
            unit_health, unit_weapon_cooldowns, idx = carry
            unit_health = unit_health.at[delta.attacked_idx].set(
                jnp.maximum(unit_health[delta.attacked_idx] + delta.health_diff, 0.0)
            )
            unit_weapon_cooldowns = unit_weapon_cooldowns.at[idx].set(
                state.unit_weapon_cooldowns[idx] + delta.cooldown_diff
            )
            return (unit_health, unit_weapon_cooldowns, idx + 1), None

        # TODO, just sum the deltas and take the max w/ 0
        (unit_health, unit_weapon_cooldowns, _), _ = jax.lax.scan(
            update_health, (state.unit_health, state.unit_weapon_cooldowns, 0), deltas
        )
        state = state.replace(
            unit_health=unit_health,
            unit_positions=deltas.pos,
            unit_weapon_cooldowns=unit_weapon_cooldowns,
        )
        return state

    def get_world_state(self, state: State) -> chex.Array:
        # get the features of every unit, as well as the teams that they belong to.
        def get_features(i):
            empty_features = jnp.zeros(shape=(len(self.own_features),))
            features = empty_features.at[0].set(
                state.unit_health[i] / self.unit_type_health[state.unit_types[i]]
            )
            features = features.at[1:3].set(state.unit_positions[i])
            features = features.at[3].set(state.unit_weapon_cooldowns[i])
            features = features.at[4 + state.unit_types[i]].set(1)
            return jax.lax.cond(
                state.unit_alive[i], lambda: features, lambda: empty_features
            )

        get_all_features = jax.vmap(get_features)
        unit_obs = get_all_features(jnp.arange(self.num_agents)).reshape(-1)
        unit_teams = state.unit_teams
        unit_types = state.unit_types
        return jnp.concatenate([unit_obs, unit_teams, unit_types], axis=-1)

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        """Applies observation function to state."""
        actions = state.prev_actions

        def get_features(i, j):
            """Get features of unit j as seen from unit i"""
            # Can just keep them symmetrical for now.
            # j here means 'the jth unit that is not i'
            # The observation is such that allies are always first
            # so for units in the second team we count in reverse.
            team_i_idx = (i >= self.num_allies).astype(jnp.int32)
            j = jax.lax.cond(
                i < self.num_allies,
                lambda: j,
                lambda: self.num_agents - j - 1,
            )
            offset = jax.lax.cond(i < self.num_allies, lambda: 1, lambda: -1)
            j_idx = jax.lax.cond(
                ((j < i) & (i < self.num_allies)) | ((j > i) & (i >= self.num_allies)),
                lambda: j,
                lambda: j + offset,
            )
            team_j_idx = (j_idx >= self.num_allies).astype(jnp.int32)
            empty_features = jnp.zeros(shape=(len(self.unit_features),))
            features = empty_features.at[0].set(
                state.unit_health[j_idx]
                / self.unit_type_health[state.unit_types[j_idx]]
            )
            features = features.at[1:3].set(
                (state.unit_positions[j_idx] - state.unit_positions[i])
                / self.unit_type_sight_ranges[state.unit_types[i]]
            )
            # TODO encode as one hot?
            action_obs = jax.lax.select(
                (team_i_idx == team_j_idx) | self.see_enemy_actions, actions[j_idx], 0
            )
            features = features.at[3].set(action_obs)
            features = features.at[4].set(state.unit_weapon_cooldowns[j_idx])
            features = features.at[5 + state.unit_types[j_idx]].set(1)
            visible = (
                jnp.linalg.norm(state.unit_positions[j_idx] - state.unit_positions[i])
                < self.unit_type_sight_ranges[state.unit_types[i]]
            )
            return jax.lax.cond(
                visible & state.unit_alive[i] & state.unit_alive[j_idx],
                lambda: features,
                lambda: empty_features,
            )

        def get_self_features(i):
            empty_features = jnp.zeros(shape=(len(self.own_features),))
            features = empty_features.at[0].set(
                state.unit_health[i] / self.unit_type_health[state.unit_types[i]]
            )
            features = features.at[1:3].set(
                state.unit_positions[i] / jnp.array([self.map_width, self.map_height])
            )
            features = features.at[3].set(state.unit_weapon_cooldowns[i])
            features = features.at[4 + state.unit_types[i]].set(1)
            return jax.lax.cond(
                state.unit_alive[i], lambda: features, lambda: empty_features
            )

        get_all_features_for_unit = jax.vmap(get_features, in_axes=(None, 0))
        get_all_features = jax.vmap(get_all_features_for_unit, in_axes=(0, None))
        other_unit_obs = get_all_features(
            jnp.arange(self.num_agents), jnp.arange(self.num_agents - 1)
        )
        other_unit_obs = other_unit_obs.reshape((self.num_agents, -1))
        get_all_self_features = jax.vmap(get_self_features)
        own_unit_obs = get_all_self_features(jnp.arange(self.num_agents))
        obs = jnp.concatenate([other_unit_obs, own_unit_obs], axis=-1)
        return {agent: obs[self.agent_ids[agent]] for agent in self.agents}

    @partial(jax.jit, static_argnums=(0,))
    def get_avail_actions(self, state: State) -> Dict[str, chex.Array]:
        @partial(jax.jit, static_argnums=(1,))
        def get_individual_avail_actions(i, team):
            num_actions = {0: self.num_ally_actions, 1: self.num_enemy_actions}[team]
            is_alive = state.unit_alive[i]
            mask = jnp.zeros((num_actions,), dtype=jnp.uint8)
            # always can take the stop action
            mask = mask.at[self.num_movement_actions - 1].set(1)
            mask = mask.at[: self.num_movement_actions - 1].set(
                jax.lax.select(
                    is_alive,
                    jnp.ones((self.num_movement_actions - 1,), dtype=jnp.uint8),
                    jnp.zeros((self.num_movement_actions - 1,), dtype=jnp.uint8),
                )
            )
            shootable_mask = (
                jnp.linalg.norm(state.unit_positions - state.unit_positions[i], axis=-1)
                < self.unit_type_attack_ranges[state.unit_types[i]]
            ) & state.unit_alive
            shootable_mask = shootable_mask if team == 0 else shootable_mask[::-1]
            shootable_mask = (
                shootable_mask[self.num_allies :]
                if team == 0
                else shootable_mask[self.num_enemies :]
            )
            shootable_mask = jax.lax.select(
                is_alive, shootable_mask, jnp.zeros_like(shootable_mask)
            )
            mask = mask.at[self.num_movement_actions :].set(shootable_mask)
            return mask

        ally_avail_actions_masks = jax.vmap(
            get_individual_avail_actions, in_axes=(0, None)
        )(jnp.arange(self.num_allies), 0)
        enemy_avail_actions_masks = jax.vmap(
            get_individual_avail_actions, in_axes=(0, None)
        )(jnp.arange(self.num_allies, self.num_agents), 1)
        return {
            agent: ally_avail_actions_masks[i]
            if i < self.num_allies
            else enemy_avail_actions_masks[i - self.num_allies]
            for i, agent in enumerate(self.agents)
        }

    def expand_state_seq(self, state_seq):
        expanded_state_seq = []
        for key, state, actions in state_seq:
            agents = self.agents
            for _ in range(self.world_steps_per_env_step):
                expanded_state_seq.append((key, state, actions))
                world_actions = jnp.array([actions[i] for i in agents])
                key, step_key = jax.random.split(key)
                state = self._world_step(step_key, state, world_actions)
                state = self._kill_agents_touching_walls(state)
                state = self._update_dead_agents(state)
                state = self._push_units_away(state)
            state = state.replace(terminal=self.is_terminal(state))
        return expanded_state_seq

    def init_render(
        self,
        ax,
        state: Tuple[State, Dict],
        step: int,
        env_step: int,
    ):
        from matplotlib.patches import Circle, Rectangle
        import matplotlib.pyplot as plt
        import numpy as np

        _, state, actions = state

        # work out which agents are being shot
        def agent_being_shot(shooter_idx, action):
            attacked_idx = jax.lax.cond(
                shooter_idx < self.num_allies,
                lambda: action + self.num_allies - self.num_movement_actions,
                lambda: self.num_allies - 1 - (action - self.num_movement_actions),
            )
            return attacked_idx

        def agent_can_shoot(shooter_idx, action):
            attacked_idx = agent_being_shot(shooter_idx, action)
            dist = jnp.linalg.norm(
                state.unit_positions[shooter_idx] - state.unit_positions[attacked_idx]
            )
            return (
                state.unit_alive[shooter_idx]
                & state.unit_alive[attacked_idx]
                & (dist < self.unit_type_attack_ranges[state.unit_types[shooter_idx]])
            )

        attacked_agents = set(
            int(agent_being_shot(i, actions[agent]))
            for i, agent in enumerate(self.agents)
            if actions[agent] > self.num_movement_actions - 1
            and agent_can_shoot(i, actions[agent])
        )
        # render circles
        ax.clear()
        ax.set_xlim([0.0, self.map_width])
        ax.set_ylim([0.0, self.map_height])
        ax.set_title(f"Step {env_step}")
        for i in range(self.num_allies):
            if state.unit_alive[i]:
                color = "blue" if i not in attacked_agents else "cornflowerblue"
                c = Circle(
                    state.unit_positions[i],
                    self.unit_type_radiuses[state.unit_types[i]],
                    color=color,
                )
                ax.add_patch(c)
                ax.text(
                    state.unit_positions[i][0]
                    - (1.0 / jnp.sqrt(2))
                    * self.unit_type_radiuses[state.unit_types[i]],
                    state.unit_positions[i][1]
                    - (1.0 / jnp.sqrt(2))
                    * self.unit_type_radiuses[state.unit_types[i]],
                    self.unit_type_shorthands[state.unit_types[i]],
                    fontsize="xx-small",
                    color="white",
                )
        for i in range(self.num_enemies):
            idx = i + self.num_allies
            if state.unit_alive[idx]:
                color = "green" if idx not in attacked_agents else "limegreen"
                c = Circle(
                    state.unit_positions[idx],
                    self.unit_type_radiuses[state.unit_types[idx]],
                    color=color,
                )
                ax.add_patch(c)
                ax.text(
                    state.unit_positions[idx][0]
                    - (1.0 / jnp.sqrt(2))
                    * self.unit_type_radiuses[state.unit_types[idx]],
                    state.unit_positions[idx][1]
                    - (1.0 / jnp.sqrt(2))
                    * self.unit_type_radiuses[state.unit_types[idx]],
                    self.unit_type_shorthands[state.unit_types[idx]],
                    fontsize="xx-small",
                    color="white",
                )

        # render bullets
        for agent in self.agents:
            i = self.agent_ids[agent]
            attacked_idx = agent_being_shot(i, actions[agent])
            if actions[agent] < self.num_movement_actions or not agent_can_shoot(
                i, actions[agent]
            ):
                continue
            frac = step / self.world_steps_per_env_step
            bullet_pos = (1 - frac) * state.unit_positions[
                i
            ] + frac * state.unit_positions[attacked_idx]
            r = Rectangle(bullet_pos, 0.5, 0.5, color="gray")
            ax.add_patch(r)

        with io.BytesIO() as buff:
            ax.figure.savefig(buff, format="raw")
            buff.seek(0)
            data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
        w, h = ax.figure.canvas.get_width_height()
        im = data.reshape((w, h, -1))

        return ax.imshow(im)

    def update_render(
        self,
        im,
        state: State,
        step: int,
        env_step: int,
    ):
        ax = im.axes
        return self.init_render(ax, state, step, env_step)



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from functools import partial
from jaxmarl.environments.mpe.simple import (
    SimpleMPE,
    State,
    AGENT_COLOUR,
    ADVERSARY_COLOUR,
    OBS_COLOUR,
)
from jaxmarl.environments.mpe.default_params import *
from gymnax.environments.spaces import Box, Discrete


# NOTE food and forests are part of world.landmarks


class SimpleWorldCommMPE(SimpleMPE):
    """
    JAX Compatible version of simple_world_comm_v2 PettingZoo environment.
    Source code: https://github.com/Farama-Foundation/PettingZoo/blob/master/pettingzoo/mpe/simple_world_comm/simple_world_comm.py
    Note, currently only have continuous actions implemented.
    """

    def __init__(
        self,
        num_good_agents=2,
        num_adversaries=4,
        num_obs=1,
        num_food=2,
        num_forests=2,
        action_type=CONTINUOUS_ACT,
    ):
        # Fixed parameters
        dim_c = 4  # communication channel dimension

        # Number of entities in each entity class.
        num_agents = num_good_agents + num_adversaries
        num_landmarks = num_obs + num_food + num_forests

        # Number of agents in each agent class.
        self.num_good_agents, self.num_adversaries = num_good_agents, num_adversaries
        self.num_obs, self.num_food, self.num_forests = num_obs, num_food, num_forests

        # Entity names
        self.leader = "leadadversary_0"
        self.adversaries = [
            "adversary_{}".format(i) for i in range(num_adversaries - 1)
        ]
        self.good_agents = ["agent_{}".format(i) for i in range(num_good_agents)]
        agents = [self.leader] + self.adversaries + self.good_agents

        landmarks = (
            ["landmark {}".format(i) for i in range(num_obs)]
            + ["food {}".format(i) for i in range(num_food)]
            + ["forest {}".format(i) for i in range(num_forests)]
        )

        self.leader_map = jnp.insert(jnp.zeros((num_agents - 1)), 0, 1)
        self.leader_idx = 0

        # Action and observation spaces
        if action_type == DISCRETE_ACT:
            action_spaces = {i: Discrete(5) for i in agents}
            action_spaces[self.leader] = Discrete(20)
        elif action_type == CONTINUOUS_ACT:
            action_spaces = {i: Box(0.0, 1.0, (5,)) for i in agents}
            action_spaces[self.leader] = Box(0.0, 1.0, (9,))
        else:
            raise NotImplementedError("Action type not implemented")

        observation_spaces = {
            i: Box(-jnp.inf, jnp.inf, (34,)) for i in self.adversaries + [self.leader]
        }
        observation_spaces.update(
            {i: Box(-jnp.inf, jnp.inf, (28,)) for i in self.good_agents}
        )

        colour = (
            [(115, 40, 40)]
            + [ADVERSARY_COLOUR] * (num_adversaries - 1)
            + [AGENT_COLOUR] * num_good_agents
            + [OBS_COLOUR] * num_obs
            + [(39, 39, 166)] * num_food
            + [(153, 230, 153)] * num_forests
        )

        # Parameters
        rad = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 0.075),
                jnp.full((self.num_good_agents), 0.045),
                jnp.full((self.num_obs), 0.2),
                jnp.full((self.num_food), 0.03),
                jnp.full((self.num_forests), 0.3),
            ]
        )
        silent = jnp.insert(jnp.ones((num_agents - 1)), 0, 0).astype(jnp.int32)
        collide = jnp.concatenate(
            [
                jnp.full((num_agents + self.num_obs), True),
                jnp.full(self.num_food + self.num_forests, False),
            ]
        )
        accel = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 3.0),
                jnp.full((self.num_good_agents), 4.0),
            ]
        )
        max_speed = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 1.0),
                jnp.full((self.num_good_agents), 1.3),
                jnp.full((num_landmarks), 0.0),
            ]
        )

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            action_spaces=action_spaces,
            observation_spaces=observation_spaces,
            dim_c=dim_c,
            colour=colour,
            rad=rad,
            silent=silent,
            collide=collide,
            accel=accel,
            max_speed=max_speed,
        )

    def set_actions(self, actions: dict):
        """Extract actions for each agent from their action array."""
        return self.action_decoder(None, actions)

    def _decode_discrete_action(
        self, a_idx: int, actions: chex.Array
    ) -> Tuple[chex.Array, chex.Array]:
        @partial(jax.vmap, in_axes=[0, 0])
        def u_decoder(a_idx, a):
            u = jnp.zeros((self.dim_p))
            idx = jax.lax.select(a <= 2, 0, 1)
            u_val = jax.lax.select(a % 2 == 0, 1.0, -1.0) * (a != 0)
            u = u.at[idx].set(u_val)
            u = u * self.accel[a_idx] * self.moveable[a_idx]
            return u

        lact = actions[self.leader]
        aact = jnp.array([actions[a] for a in self.adversaries])
        gact = jnp.array([actions[a] for a in self.good_agents])
        luact = lact % 5  # BUG
        u_acts = jnp.concatenate((luact[None], aact, gact))
        u = u_decoder(self.agent_range, u_acts)

        c = jnp.zeros(
            (
                self.num_agents,
                self.dim_c,
            )
        )
        lcact = lact // 5
        c = c.at[0, lcact].set(1.0)
        return u, c

    def _decode_continuous_action(
        self, a_idx: int, actions: chex.Array
    ) -> Tuple[chex.Array, chex.Array]:
        @partial(jax.vmap, in_axes=[0, 0])
        def _set_u(a_idx, action):
            u = jnp.array([action[1] - action[2], action[3] - action[4]])
            return u * self.accel[a_idx] * self.moveable[a_idx]

        lact = actions[self.leader]
        aact = jnp.array([actions[a] for a in self.adversaries])
        gact = jnp.array([actions[a] for a in self.good_agents])

        u_acts = jnp.concatenate([lact[:5][None], aact, gact])
        u = _set_u(self.agent_range, u_acts)

        c = jnp.zeros((self.num_agents, self.dim_c))
        c = c.at[self.leader_idx].set(lact[5:])
        return u, c

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        """Returns observations of all agents"""

        @partial(jax.vmap, in_axes=(0, None))
        def _in_forest(idx: int, state: State) -> chex.Array:
            """Collision check for all forests with agent `idx`"""
            dist = jnp.linalg.norm(
                state.p_pos[self.num_agents + self.num_obs + self.num_food :]
                - state.p_pos[idx],
                axis=1,
            )
            dist_min = self.rad[-self.num_forests :] + self.rad[idx]
            return dist < dist_min

        @partial(jax.vmap, in_axes=(0, None, None))
        def _common_stats(aidx: int, forest: chex.Array, state: State):
            """Values needed in all observations"""

            landmark_pos = (
                state.p_pos[self.num_agents :] - state.p_pos[aidx]
            )  # Landmark positions in agent reference frame

            in_forest = jnp.any(forest[aidx])  # True if ego agent in forest
            same_forest = jnp.any(
                forest[aidx] * forest, axis=1
            )  # True if other and ego agent in same forest
            no_forest = (
                jnp.all(~forest, axis=1) & ~in_forest
            )  # True if other not in a forest and ego agent also not in a forest

            leader = aidx == self.leader_idx
            other_mask = (
                jnp.logical_or(same_forest, no_forest) | leader
            )  # Whether ego agent can see other agent

            # Zero out unseen agents with other_mask
            other_pos = (
                state.p_pos[: self.num_agents] - state.p_pos[aidx]
            ) * other_mask[:, None]
            other_vel = state.p_vel[: self.num_agents] * other_mask[:, None]

            # use jnp.roll to remove ego agent from other_pos and other_vel arrays
            other_pos = jnp.roll(other_pos, shift=self.num_agents - aidx - 1, axis=0)[
                : self.num_agents - 1
            ]
            other_vel = jnp.roll(other_vel, shift=self.num_agents - aidx - 1, axis=0)[
                : self.num_agents - 1
            ]

            other_pos = jnp.roll(other_pos, shift=aidx, axis=0)
            other_vel = jnp.roll(other_vel, shift=aidx, axis=0)

            return landmark_pos, other_pos, other_vel, jnp.where(forest[aidx], 1, -1)

        forest = _in_forest(self.agent_range, state)
        landmark_pos, other_pos, other_vel, forest = _common_stats(
            self.agent_range, forest, state
        )

        # NOTE orderings taken from MPE code and some differ to their docs
        def _good(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                    forest[aidx],  # 2
                    other_vel[aidx, -1:].flatten(),  # 2
                ]
            )

        def _leader():
            return jnp.concatenate(
                [
                    state.p_vel[self.leader_idx][None].flatten(),
                    state.p_pos[self.leader_idx][None].flatten(),
                    landmark_pos[self.leader_idx].flatten(),
                    other_pos[self.leader_idx].flatten(),
                    other_vel[self.leader_idx, -2:].flatten(),  # 2, 2
                    forest[self.leader_idx],  # NOTE this differs to their docs
                    state.c[self.leader_idx][None].flatten(),  # 4
                ]
            )

        def _adversary(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx][None].flatten(),
                    state.p_pos[aidx][None].flatten(),
                    landmark_pos[aidx].flatten(),
                    other_pos[aidx].flatten(),
                    other_vel[aidx, -2:].flatten(),
                    forest[aidx],
                    state.c[self.leader_idx][None].flatten(),
                ]
            )

        # Format observations as a dictionary keyed by agent name
        obs = {self.leader: _leader()}
        obs.update({a: _adversary(i + 1) for i, a in enumerate(self.adversaries)})
        obs.update(
            {a: _good(i + self.num_adversaries) for i, a in enumerate(self.good_agents)}
        )
        return obs

    def rewards(self, state: State) -> Dict[str, float]:
        """Computes rewards for all agents"""

        @partial(jax.vmap, in_axes=[0, None])
        def _reward(aidx: int, state: State):
            return jax.lax.cond(
                aidx < self.num_adversaries,
                self.adversary_reward,
                self.agent_reward,
                *(aidx, state)
            )

        r = _reward(self.agent_range, state)
        return {agent: r[i] for i, agent in enumerate(self.agents)}

    def agent_reward(self, aidx: int, state: State):
        """Reward for good agents."""

        @partial(jax.vmap, in_axes=(0,))
        def _bound_rew(x):
            w = x < 0.9
            m = x < 1.0
            mr = (x - 0.9) * 10
            br = jnp.min(jnp.array([jnp.exp(2 * x - 2), 10]))

            return jax.lax.select(m, mr, br) * ~w

        rew = 0
        # check collision, -5 for each collision with adversary
        ac = self._collision(
            state.p_pos[aidx],
            self.rad[aidx],
            state.p_pos[: self.num_adversaries],
            self.rad[: self.num_adversaries],
        )
        rew -= jnp.sum(ac) * 5

        # check map bounds,
        rew -= 2 * jnp.sum(_bound_rew(jnp.abs(state.p_pos[aidx])))

        # check food collisions
        fc = self._collision(
            state.p_pos[aidx],
            self.rad[aidx],
            state.p_pos[-(self.num_food + self.num_forests) : -self.num_forests],
            self.rad[-(self.num_food + self.num_forests) : -self.num_forests],
        )
        rew += jnp.sum(fc) * 2

        # reward for being near food
        rew -= 0.05 * jnp.min(
            jnp.linalg.norm(
                state.p_pos[-(self.num_food + self.num_forests) : -self.num_forests]
                - state.p_pos[aidx],
                axis=1,
            )
        )
        return rew

    def adversary_reward(self, aidx: int, state: State):
        """Reward for adversary agents."""

        @partial(jax.vmap, in_axes=[0, 0, None, None])
        def vcollision(apos, arad, opos, orad):
            return self._collision(apos, arad, opos, orad)

        rew = 0

        rew -= 0.1 * jnp.min(
            jnp.linalg.norm(
                state.p_pos[self.num_adversaries : self.num_agents] - state.p_pos[aidx],
                axis=1,
            )
        )

        # for each agent, add collision bonus
        rew += 5 * jnp.sum(
            vcollision(
                state.p_pos[self.num_adversaries : self.num_agents],
                self.rad[self.num_adversaries : self.num_agents],
                state.p_pos[: self.num_adversaries],
                self.rad[: self.num_adversaries],
            )
        )
        return rew

    @partial(jax.vmap, in_axes=(None, None, None, 0, 0))
    def _collision(self, apos, arad, opos, orad):
        """Check collision between two entities."""
        deltas = opos - apos
        size = arad + orad
        dist = jnp.sqrt(jnp.sum(deltas**2))
        return dist < size


def test_policy(key, state: State):
    """Test policy where the adversaries hunt the first good agent"""
    pos = state.p_pos[3]

    act = jnp.zeros((5, 9))

    o = pos - state.p_pos[:3]
    act = act.at[:3, 1].set(o[:, 0])
    act = act.at[:3, 3].set(o[:, 1])

    r = jax.random.uniform(key, (2, 9))
    act = act.at[3:].set(r)
    return act


if __name__ == "__main__":
    from pettingzoo.mpe import simple_world_comm_v3

    ### Petting zoo env
    zoo_env = simple_world_comm_v3.parallel_env(max_cycles=25, continuous_actions=True)
    zoo_obs = zoo_env.reset()
    actions = {agent: zoo_env.action_space(agent).sample() for agent in zoo_env.agents}

    obs_space = {agent: zoo_env.observation_space(agent) for agent in zoo_env.agents}
    act_space = {agent: zoo_env.action_space(agent) for agent in zoo_env.agents}
    print("obs space", obs_space, "\n act space", act_space)
    # print('zoo obs', zoo_obs)
    key = jax.random.PRNGKey(0)

    env = SimpleWorldCommMPE()

    key, key_r = jax.random.split(key)
    obs, state = env.reset(key_r)

    # obs = env.observation(0, state)
    # print('obs', obs.shape, obs)

    mock_action = jnp.array([[0.0, 0.0, 1.0, 0.1, 0.1, 0.0, 0.0, 0.0, 0.0]])
    #
    # actions = jnp.repeat(mock_action[None], repeats=env.num_agents, axis=0).squeeze()
    # actions = {agent: mock_action for agent in env.agents}
    # env.enable_render()

    print("state", state)
    for _ in range(50):
        key, key_a, key_s = jax.random.split(key, 3)
        # actions = test_policy(key_a, state)
        # actions = {agent: actions[i] for i, agent in enumerate(env.agents)}
        # print('actions', actions)
        # print('state', state)
        obs, state, rew, dones, _ = env.step_env(key_s, state, actions)
        actions = {
            agent: zoo_env.action_space(agent).sample() for agent in zoo_env.agents
        }
        # env.render(state)
        print("obs", [o.shape for o in obs.values()])
        # raise
        # print('rew', rew)



from typing import Callable, Any, Tuple

import chex
import jax.lax

ReplayBufferState = Any
Item = chex.ArrayTree
ItemBatch = chex.ArrayTree
IntScalar = chex.Array
ItemUpdateFn = Callable[[Item], Item]


@chex.dataclass(frozen=True)
class ReplayBuffer:
    init_fn: Callable[[Item], ReplayBufferState]
    size_fn: Callable[[ReplayBufferState], IntScalar]
    add_fn: Callable[[ReplayBufferState, Item], ReplayBufferState]
    add_batch_fn: Callable[[ReplayBufferState, ItemBatch], ReplayBufferState]
    sample_fn: Callable[[ReplayBufferState, chex.PRNGKey, int], ItemBatch]
    update_fn: Callable[[ReplayBufferState, ItemUpdateFn], ReplayBufferState]


def make_default_add_batch_fn(add_fn):
    def add_batch_fn(state: ReplayBufferState, item_batch: ItemBatch) -> ReplayBufferState:
        def scan_body(state: ReplayBufferState, item: Item) -> Tuple[ReplayBufferState, None]:
            state = add_fn(state, item)
            return state, None

        state, _ = jax.lax.scan(f=scan_body, init=state, xs=item_batch)
        return state

    return add_batch_fn





import argparse
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np

# from jaxmarl.gridworld.maze import Maze #, Actions
# from jaxmarl.gridworld.ma_maze import MAMaze
from jaxmarl.environments.overcooked.overcooked import Overcooked
from jaxmarl.environments.overcooked.layouts import overcooked_layouts as layouts


def redraw(state, obs, extras):
    extras['viz'].render(extras['agent_view_size'], state, highlight=False)

    # if extras['obs_viz'] is not None:
    #     if extras['env'] == "MAMaze" or "Overcooked":
    #         obs_viz.render_grid(np.asarray(obs['image'][0]), k_rot90=3, agent_dir_idx=[3])
    #         obs_viz2.render_grid(np.asarray(obs['image'][1]), k_rot90=3, agent_dir_idx=[3])
    #     else:
    #         obs_viz.render_grid(np.asarray(obs['image']), k_rot90=3, agent_dir_idx=3)


def reset(key, env, extras):
    key, subkey = jax.random.split(extras['rng'])
    obs, state = extras['jit_reset'](subkey)

    extras['rng'] = key
    extras['obs'] = obs
    extras['state'] = state

    redraw(state, obs, extras)


def step(env, action, extras):
    key, subkey = jax.random.split(extras['rng'])

    actions = {"agent_0" : jnp.array(action), "agent_1" : jnp.array(action)}
    print("Actions : ", actions)
    obs, state, reward, done, info = jax.jit(env.step_env)(subkey, extras['state'], actions)
    extras['obs'] = obs
    extras['state'] = state
    print(f"t={state.time}: reward={reward['agent_0']}, agent_dir={state.agent_dir_idx}, agent_inv={state.agent_inv}, done = {done['__all__']}")
    
    if extras["debug"]:
        layers = [f"player_{i}_loc" for i in range(2)]
        layers.extend([f"player_{i // 4}_orientation_{i % 4}" for i in range(8)])
        layers.extend([
            "pot_loc",
            "counter_loc",
            "onion_disp_loc",
            "tomato_disp_loc",
            "plate_disp_loc",
            "serve_loc",
            "onions_in_pot",
            "tomatoes_in_pot",
            "onions_in_soup",
            "tomatoes_in_soup",
            "soup_cook_time_remaining",
            "soup_done",
            "plates",
            "onions",
            "tomatoes",
            "urgency"
        ])
        print("obs_shape: ", obs["agent_0"].shape)
        print("OBS: \n", obs["agent_0"])
        debug_obs = jnp.transpose(obs["agent_0"], (2,0,1))
        for i, layer in enumerate(layers):
            print(layer)
            print(debug_obs[i])
    # print(f"agent obs =\n {obs}")

    if done["__all__"] or (jnp.array([action, action]) == Actions.done).any():
        key, subkey = jax.random.split(subkey)
        reset(subkey, env, extras)
    else:
        redraw(state, obs, extras)

    extras['rng'] = key


def key_handler(env, extras, event):
    print('pressed', event.key)

    if event.key == 'escape':
        window.close()
        return

    if event.key == 'backspace':
        extras['jit_reset']((env, extras))
        return

    if event.key == 'left':
        step(env, Actions.left, extras)
        return
    if event.key == 'right':
        step(env, Actions.right, extras)
        return
    if event.key == 'up':
        step(env, Actions.forward, extras)
        return

    # Spacebar
    if event.key == ' ':
        step(env, Actions.toggle, extras)
        return
    if event.key == '[':
        step(env, Actions.pickup, extras)
        return
    if event.key == ']':
        step(env, Actions.drop, extras)
        return

    if event.key == 'enter':
        step(env, Actions.done, extras)
        return

def key_handler_overcooked(env, extras, event):
    print('pressed', event.key)

    if event.key == 'escape':
        window.close()
        return
    if event.key == 'backspace':
        extras['jit_reset']((env, extras))
        return

    if event.key == 'left':
        step(env, Actions.left, extras)
        return
    if event.key == 'right':
        step(env, Actions.right, extras)
        return
    if event.key == 'up':
        # step(env, Actions.forward, extras)
        step(env, Actions.up, extras)
        return
    if event.key == 'down':
        step(env, Actions.down, extras)
        return

    # Spacebar
    if event.key == ' ':
        step(env, Actions.interact, extras)
        return
    if event.key == 'tab':
        step(env, Actions.stay, extras)
        return
    if event.key == 'enter':
        step(env, Actions.done, extras)
        return


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--env",
        type=str,
        help="Environment name",
        default="Overcooked"
    )
    parser.add_argument(
        "--layout",
        type=str,
        help="Overcooked layout",
        default="cramped_room"
    )
    parser.add_argument(
        '--random_reset',
        default=False,
        help="Reset to random state",
        action='store_true'
    )
    parser.add_argument(
        "--seed",
        type=int,
        help="random seed to generate the environment with",
        default=0
    )
    parser.add_argument(
        '--render_agent_view',
        default=False,
        help="draw the agent sees (partially observable view)",
        action='store_true'
    )
    # parser.add_argument(
    #     '--height',
    #     default=13,
    #     type=int,
    #     help="height",
    # )
    # parser.add_argument(
    #     '--width',
    #     default=13,
    #     type=int,
    #     help="width",
    # )
    # parser.add_argument(
    #     '--n_walls',
    #     default=50,
    #     type=int,
    #     help="Number of walls",
    # )
    # parser.add_argument(
    #     '--agent_view_size',
    #     default=5,
    #     type=int,
    #     help="Number of walls",
    # )
    parser.add_argument(
        '--debug',
        default=False,
        help="Debug mode",
        action='store_true'
    )
    args = parser.parse_args()

    # if args.env == "Maze":
    #     env = Maze(
    #         height=13,
    #         width=13,
    #         n_walls=25,
    #         see_agent=True,
    #     )
    #     from jaxmarl.gridworld.grid_viz import GridVisualizer as Visualizer
    #     from jaxmarl.gridworld.maze import Actions
    #
    #     params = env.params
    #
    # elif args.env == "MAMaze":
    #     env = MAMaze(
    #         height=13,
    #         width=13,
    #         n_walls=25,
    #         see_agent=True,
    #         n_agents=2
    #     )
    #     from jaxmarl.gridworld.grid_viz import GridVisualizer as Visualizer
    #     from jaxmarl.gridworld.maze import Actions
    #
    #     params = env.params

    if args.env == "Overcooked":
        if len(args.layout) > 0:
            layout = layouts[args.layout]
            env = Overcooked(
                layout=layout,
                random_reset=args.random_reset
            )
        else:
            print("You must provide a layout.")
        from jaxmarl.viz.overcooked_visualizer import OvercookedVisualizer as Visualizer
        from jaxmarl.environments.overcooked.overcooked import Actions


    viz = Visualizer()
    obs_viz = None
    obs_viz2 = None
    if args.render_agent_view:
        obs_viz = Visualizer()
        if args.env == "MAMaze" or "Overcooked":
            obs_viz2 = Visualizer()

    with jax.disable_jit(False):
        jit_reset = jax.jit(env.reset)
        # jit_reset = env.reset_env
        key = jax.random.PRNGKey(args.seed)
        key, subkey = jax.random.split(key)
        o0, s0 = jit_reset(subkey)
        viz.render(env.agent_view_size, s0, highlight=False)

        key, subkey = jax.random.split(key)
        extras = {
            'rng': subkey,
            'state': s0,
            'obs': o0,
            'viz': viz,
            'obs_viz': obs_viz,
            'obs_viz2': obs_viz2,
            'jit_reset': jit_reset,
            'agent_view_size': env.agent_view_size,
            'env': args.env,
            'debug': args.debug
        }

        if args.env == "Overcooked":
            viz.window.reg_key_handler(partial(key_handler_overcooked, env, extras))
            viz.show(block=True)
        else:
            viz.window.reg_key_handler(partial(key_handler, env, extras))
            viz.show(block=True)




import chex
import jax
import jax.experimental.checkify as checkify

from . import circular_buffer
from .base import ReplayBuffer, Item, ItemBatch, IntScalar, ItemUpdateFn, make_default_add_batch_fn


@chex.dataclass(frozen=True)
class UniformReplayBufferState:
    storage: circular_buffer.CircularBuffer


def uniform_sample(
        buffer: circular_buffer.CircularBuffer, rng: chex.PRNGKey, batch_size: int
) -> circular_buffer.ItemBatch:
    checkify.check(circular_buffer.size(buffer) > 0, 'Cannot sample from an empty buffer')

    sample_pos = jax.random.randint(rng, minval=0, maxval=circular_buffer.size(buffer), shape=(batch_size,))
    get_at_index_batch = jax.vmap(circular_buffer.get_at_index, in_axes=(None, 0))
    return get_at_index_batch(buffer, sample_pos)


def uniform_replay(max_size: int) -> ReplayBuffer:
    def init_fn(item_prototype: Item) -> UniformReplayBufferState:
        return UniformReplayBufferState(storage=circular_buffer.init(item_prototype, max_size))

    def size_fn(state: UniformReplayBufferState) -> IntScalar:
        return circular_buffer.size(state.storage)

    def add_fn(state: UniformReplayBufferState, item: Item) -> UniformReplayBufferState:
        return state.replace(storage=circular_buffer.push(state.storage, item))

    def sample_fn(state: UniformReplayBufferState, rng: chex.PRNGKey, batch_size: int) -> ItemBatch:
        return uniform_sample(state.storage, rng, batch_size)

    def update_fn(state: UniformReplayBufferState, item_update_fn: ItemUpdateFn) -> UniformReplayBufferState:
        # TODO: there might be a faster way to make updates that does not affect all items in the buffer
        batch_update_fn = jax.vmap(item_update_fn)
        updated_data = batch_update_fn(state.storage.data)
        return state.replace(storage=state.storage.replace(data=updated_data))

    return ReplayBuffer(
        init_fn=jax.tree_util.Partial(init_fn),
        size_fn=jax.tree_util.Partial(size_fn),
        add_fn=jax.tree_util.Partial(add_fn),
        # TODO: it should be possible to make an optimized version of add_batch_fn for this buffer type
        add_batch_fn=jax.tree_util.Partial(make_default_add_batch_fn(add_fn)),
        sample_fn=jax.tree_util.Partial(sample_fn),
        update_fn=jax.tree_util.Partial(update_fn),
    )



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from functools import partial
from jaxmarl.environments.mpe.simple import State, SimpleMPE
from jaxmarl.environments.mpe.default_params import *
from gymnax.environments.spaces import Box


class SimpleAdversaryMPE(SimpleMPE):
    def __init__(
        self,
        num_good_agents=2,
        num_adversaries=1,
        num_obs=2,
        action_type=DISCRETE_ACT,
    ):
        dim_c = 2  # NOTE follows code rather than docs

        num_agents = num_good_agents + num_adversaries
        num_landmarks = num_obs
        num_entities = num_landmarks + num_agents

        self.num_good_agents, self.num_adversaries = num_good_agents, num_adversaries

        self.adversaries = ["adversary_{}".format(i) for i in range(num_adversaries)]
        self.good_agents = ["agent_{}".format(i) for i in range(num_good_agents)]
        agents = self.adversaries + self.good_agents

        landmarks = ["landmark {}".format(i) for i in range(num_obs)]

        # Action and observation spaces
        observation_spaces = {i: Box(-jnp.inf, jnp.inf, (8,)) for i in self.adversaries}
        observation_spaces.update(
            {i: Box(-jnp.inf, jnp.inf, (10,)) for i in self.good_agents}
        )

        colour = (
            [ADVERSARY_COLOUR] * num_adversaries
            + [AGENT_COLOUR] * num_good_agents
            + [OBS_COLOUR] * num_obs
        )

        # Parameters
        rad = jnp.concatenate(
            [jnp.full((num_agents), 0.15), jnp.full((num_landmarks), 0.08)]
        )
        collide = jnp.full((num_entities), False)

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            observation_spaces=observation_spaces,
            dim_c=dim_c,
            colour=colour,
            rad=rad,
            collide=collide,
        )

    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        key_a, key_l, key_g = jax.random.split(key, 3)

        p_pos = jnp.concatenate(
            [
                jax.random.uniform(key_a, (self.num_agents, 2), minval=-1, maxval=+1),
                jax.random.uniform(
                    key_l, (self.num_landmarks, 2), minval=-1.0, maxval=+1.0
                ),
            ]
        )

        g_idx = jax.random.randint(key_g, (), minval=0, maxval=self.num_landmarks)

        state = State(
            p_pos=p_pos,
            p_vel=jnp.zeros((self.num_entities, self.dim_p)),
            c=jnp.zeros((self.num_agents, self.dim_c)),
            done=jnp.full((self.num_agents), False),
            step=0,
            goal=g_idx,
        )

        return self.get_obs(state), state

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        @partial(jax.vmap, in_axes=(0, None))
        def _common_stats(aidx, state: State):
            """Values needed in all observations"""

            landmark_pos = (
                state.p_pos[self.num_agents :] - state.p_pos[aidx]
            )  # Landmark positions in agent reference frame

            # Zero out unseen agents with other_mask
            other_pos = state.p_pos[: self.num_agents] - state.p_pos[aidx]

            # use jnp.roll to remove ego agent from other_pos and other_vel arrays
            other_pos = jnp.roll(other_pos, shift=self.num_agents - aidx - 1, axis=0)[
                : self.num_agents - 1
            ]

            other_pos = jnp.roll(other_pos, shift=aidx, axis=0)

            return landmark_pos, other_pos

        landmark_pos, other_pos = _common_stats(self.agent_range, state)

        def _good(aidx):
            goal_rel_pos = state.p_pos[state.goal + self.num_agents] - state.p_pos[aidx]

            return jnp.concatenate(
                [
                    goal_rel_pos.flatten(),
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                ]
            )

        def _adversary(aidx):
            return jnp.concatenate(
                [
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                ]
            )

        obs = {a: _adversary(i) for i, a in enumerate(self.adversaries)}
        obs.update(
            {a: _good(i + self.num_adversaries) for i, a in enumerate(self.good_agents)}
        )
        return obs

    def rewards(
        self,
        state: State,
    ) -> Dict[str, float]:
        adv_rew = jnp.sum(
            jnp.linalg.norm(
                state.p_pos[: self.num_adversaries]
                - state.p_pos[state.goal + self.num_agents],
                axis=1,
            )
        )
        pos_rew = -1 * jnp.min(
            jnp.linalg.norm(
                state.p_pos[self.num_adversaries : self.num_agents]
                - state.p_pos[state.goal + self.num_agents],
                axis=1,
            )
        )
        good_rew = adv_rew + pos_rew

        def _ad(aidx):
            return -1 * jnp.linalg.norm(
                state.p_pos[aidx] - state.p_pos[state.goal + self.num_agents]
            )

        rew = {a: _ad(i) for i, a in enumerate(self.adversaries)}
        rew.update({a: good_rew for a in self.good_agents})
        return rew



from setuptools import find_packages, setup
import os
import re
from typing import List

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(CURRENT_DIR, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

def _parse_requirements(path: str) -> List[str]:
    """Returns content of given requirements file."""
    with open(os.path.join(path)) as f:
        return [
            line.rstrip() for line in f if not (line.isspace() or line.startswith("#"))
        ]
        
VERSIONFILE = "jaxmarl/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))
git_tar = f"https://github.com/FLAIROx/JaxMARL/archive/v{verstr}.tar.gz"

setup(
    name="jaxmarl",
    version=verstr,
    author="Foerster Lab for AI Research",
    description="Multi-Agent Reinforcement Learning with JAX",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FLAIROx/JaxMARL",
    download_url=git_tar,
    keywords="MARL reinforcement-learning python jax",
    packages=find_packages(exclude=["baselines"]),
    python_requires=">=3.8",
    install_requires=_parse_requirements("requirements/requirements.txt"),
    extras_require={
        "dev": _parse_requirements("requirements/requirements-dev.txt"),
    },
    
)



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from functools import partial
from jaxmarl.environments.mpe.simple import SimpleMPE, State
from jaxmarl.environments.mpe.default_params import *
from gymnax.environments.spaces import Box, Discrete

# Obstacle Colours
OBS_COLOUR = [(191, 64, 64), (64, 191, 64), (64, 64, 191)]

class SimpleReferenceMPE(SimpleMPE):
    def __init__(
        self,
        num_agents=2,
        num_landmarks=3,
        local_ratio=0.5,
        action_type=DISCRETE_ACT,
    ):
        assert num_agents == 2, "SimpleReferenceMPE only supports 2 agents"
        assert num_landmarks == 3, "SimpleReferenceMPE only supports 3 landmarks"

        num_entites = num_agents + num_landmarks

        self.local_ratio = local_ratio

        dim_c = 10

        agents = ["agent_{}".format(i) for i in range(num_agents)]

        landmarks = ["landmark {}".format(i) for i in range(num_landmarks)]

        # Action and observation spaces
        if action_type == DISCRETE_ACT:
            action_spaces = {i: Discrete(50) for i in agents}
        elif action_type == CONTINUOUS_ACT:
            action_spaces = {i: Box(0.0, 1.0, (15,)) for i in agents}
        else:
            raise NotImplementedError("Action type not implemented")

        observation_spaces = {i: Box(-jnp.inf, jnp.inf, (21,)) for i in agents}
        colour = [AGENT_COLOUR] * num_agents + OBS_COLOUR

        silent = jnp.full((num_agents), 0)
        collide = jnp.full((num_entites), False)

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            action_spaces=action_spaces,
            observation_spaces=observation_spaces,
            dim_c=dim_c,
            colour=colour,
            silent=silent,
            collide=collide,
        )

    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        key_a, key_l, key_g = jax.random.split(key, 3)

        p_pos = jnp.concatenate(
            [
                jax.random.uniform(key_a, (self.num_agents, 2), minval=-1, maxval=+1),
                jax.random.uniform(
                    key_l, (self.num_landmarks, 2), minval=-1.0, maxval=+1.0
                ),
            ]
        )

        g_idx = jax.random.randint(key_g, (2,), minval=0, maxval=self.num_landmarks)

        state = State(
            p_pos=p_pos,
            p_vel=jnp.zeros((self.num_entities, self.dim_p)),
            c=jnp.zeros((self.num_agents, self.dim_c)),
            done=jnp.full((self.num_agents), False),
            step=0,
            goal=g_idx,
        )

        return self.get_obs(state), state

    def get_obs(
        self,
        state: State,
    ) -> Dict[str, chex.Array]:
        @partial(jax.vmap, in_axes=(0, None))
        def _common_stats(aidx: int, state: State):
            """Values needed in all observations"""

            landmark_pos = (
                state.p_pos[self.num_agents :] - state.p_pos[aidx]
            )  # Landmark positions in agent reference frame

            return landmark_pos

        landmark_pos = _common_stats(self.agent_range, state)

        def _agent(aidx):
            other_idx = (aidx + 1) % 2
            colour = jnp.full((3,), 0.25)
            colour = colour.at[state.goal[other_idx]].set(0.75)
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 3, 2
                    colour.flatten(),  # 3
                    state.c[other_idx].flatten(),  # 10
                ]
            )

        obs = {a: _agent(i) for i, a in enumerate(self.agents)}
        return obs

    @partial(jax.vmap, in_axes=[None, 0, 0])
    def _decode_discrete_action(
        self, a_idx: int, action: chex.Array
    ) -> Tuple[chex.Array, chex.Array]:
        u = jnp.zeros((self.dim_p,))
        c = jnp.zeros((self.dim_c,))
        u_act = action % 5
        c_act = action // 5
        idx = jax.lax.select(u_act <= 2, 0, 1)
        u_val = jax.lax.select(u_act % 2 == 0, 1.0, -1.0) * (u_act != 0)
        u = u.at[idx].set(u_val)
        u = u * self.accel[a_idx] * self.moveable[a_idx]
        c = c.at[c_act].set(1.0)
        return u, c

    def rewards(self, state: State) -> Dict[str, float]:
        @partial(jax.vmap, in_axes=(0, None))
        def _agent(aidx, state):
            other_idx = (aidx + 1) % 2
            return -1 * jnp.linalg.norm(
                state.p_pos[other_idx]
                - state.p_pos[self.num_agents + state.goal[other_idx]]
            )

        agent_rew = _agent(self.agent_range, state)
        global_rew = jnp.sum(agent_rew) / self.num_agents
        rew = {
            a: global_rew * (1 - self.local_ratio) + agent_rew[i] * self.local_ratio
            for i, a in enumerate(self.agents)
        }
        return rew



""" 
Base class for MPE PettingZoo envs.

TODO: viz for communication env, e.g. crypto
"""

import jax
import jax.numpy as jnp
import numpy as onp
from jaxmarl.environments.multi_agent_env import MultiAgentEnv
from jaxmarl.environments.mpe.default_params import *
import chex
from gymnax.environments.spaces import Box, Discrete
from flax import struct
from typing import Tuple, Optional, Dict
from functools import partial

import matplotlib.pyplot as plt
import matplotlib

@struct.dataclass
class State:
    """Basic MPE State"""

    p_pos: chex.Array  # [num_entities, [x, y]]
    p_vel: chex.Array  # [n, [x, y]]
    c: chex.Array  # communication state [num_agents, [dim_c]]
    done: chex.Array  # bool [num_agents, ]
    step: int  # current step
    goal: int = None  # index of target landmark, used in: SimpleSpeakerListenerMPE, SimpleReferenceMPE, SimplePushMPE, SimpleAdversaryMPE


class SimpleMPE(MultiAgentEnv):
    def __init__(
        self,
        num_agents=1,
        action_type=DISCRETE_ACT,
        agents=None,
        num_landmarks=1,
        landmarks=None,
        action_spaces=None,
        observation_spaces=None,
        colour=None,
        dim_c=0,
        dim_p=2,
        max_steps=MAX_STEPS,
        dt=DT,
        **kwargs,
    ):
        # Agent and entity constants
        self.num_agents = num_agents
        self.num_landmarks = num_landmarks
        self.num_entities = num_agents + num_landmarks
        self.agent_range = jnp.arange(num_agents)
        self.entity_range = jnp.arange(self.num_entities)

        # Setting, and sense checking, entity names and agent action spaces
        if agents is None:
            self.agents = [f"agent_{i}" for i in range(num_agents)]
        else:
            assert (
                len(agents) == num_agents
            ), f"Number of agents {len(agents)} does not match number of agents {num_agents}"
            self.agents = agents
        self.a_to_i = {a: i for i, a in enumerate(self.agents)}
        self.classes = self.create_agent_classes()

        if landmarks is None:
            self.landmarks = [f"landmark {i}" for i in range(num_landmarks)]
        else:
            assert (
                len(landmarks) == num_landmarks
            ), f"Number of landmarks {len(landmarks)} does not match number of landmarks {num_landmarks}"
            self.landmarks = landmarks
        self.l_to_i = {l: i + self.num_agents for i, l in enumerate(self.landmarks)}

        if action_spaces is None:
            if action_type == DISCRETE_ACT:
                self.action_spaces = {i: Discrete(5) for i in self.agents}
            elif action_type == CONTINUOUS_ACT:
                self.action_spaces = {i: Box(0.0, 1.0, (5,)) for i in self.agents}
        else:
            assert (
                len(action_spaces.keys()) == num_agents
            ), f"Number of action spaces {len(action_spaces.keys())} does not match number of agents {num_agents}"
            self.action_spaces = action_spaces

        if observation_spaces is None:
            self.observation_spaces = {
                i: Box(-jnp.inf, jnp.inf, (4,)) for i in self.agents
            }
        else:
            assert (
                len(observation_spaces.keys()) == num_agents
            ), f"Number of observation spaces {len(observation_spaces.keys())} does not match number of agents {num_agents}"
            self.observation_spaces = observation_spaces

        self.colour = (
            colour
            if colour is not None
            else [AGENT_COLOUR] * num_agents + [OBS_COLOUR] * num_landmarks
        )

        # Action type
        if action_type == DISCRETE_ACT:
            self.action_decoder = self._decode_discrete_action
        elif action_type == CONTINUOUS_ACT:
            self.action_decoder = self._decode_continuous_action
        else:
            raise NotImplementedError(f"Action type: {action_type} is not supported")

        # World dimensions
        self.dim_c = dim_c  # communication channel dimensionality
        self.dim_p = dim_p  # position dimensionality

        # Environment parameters
        self.max_steps = max_steps
        self.dt = dt
        if "rad" in kwargs:
            self.rad = kwargs["rad"]
            assert (
                len(self.rad) == self.num_entities
            ), f"Rad array length {len(self.rad)} does not match number of entities {self.num_entities}"
            assert jnp.all(self.rad > 0), f"Rad array must be positive, got {self.rad}"
        else:
            self.rad = jnp.concatenate(
                [jnp.full((self.num_agents), 0.15), jnp.full((self.num_landmarks), 0.2)]
            )

        if "moveable" in kwargs:
            self.moveable = kwargs["moveable"]
            assert (
                len(self.moveable) == self.num_entities
            ), f"Moveable array length {len(self.moveable)} does not match number of entities {self.num_entities}"
            assert (
                self.moveable.dtype == bool
            ), f"Moveable array must be boolean, got {self.moveable}"
        else:
            self.moveable = jnp.concatenate(
                [
                    jnp.full((self.num_agents), True),
                    jnp.full((self.num_landmarks), False),
                ]
            )

        if "silent" in kwargs:
            self.silent = kwargs["silent"]
            assert (
                len(self.silent) == self.num_agents
            ), f"Silent array length {len(self.silent)} does not match number of agents {self.num_agents}"
        else:
            self.silent = jnp.full((self.num_agents), 1)

        if "collide" in kwargs:
            self.collide = kwargs["collide"]
            assert (
                len(self.collide) == self.num_entities
            ), f"Collide array length {len(self.collide)} does not match number of entities {self.num_entities}"
        else:
            self.collide = jnp.full((self.num_entities), False)

        if "mass" in kwargs:
            self.mass = kwargs["mass"]
            assert (
                len(self.mass) == self.num_entities
            ), f"Mass array length {len(self.mass)} does not match number of entities {self.num_entities}"
            assert jnp.all(
                self.mass > 0
            ), f"Mass array must be positive, got {self.mass}"
        else:
            self.mass = jnp.full((self.num_entities), 1.0)

        if "accel" in kwargs:
            self.accel = kwargs["accel"]
            assert (
                len(self.accel) == self.num_agents
            ), f"Accel array length {len(self.accel)} does not match number of agents {self.num_agents}"
            assert jnp.all(
                self.accel > 0
            ), f"Accel array must be positive, got {self.accel}"
        else:
            self.accel = jnp.full((self.num_agents), 5.0)

        if "max_speed" in kwargs:
            self.max_speed = kwargs["max_speed"]
            assert (
                len(self.max_speed) == self.num_entities
            ), f"Max speed array length {len(self.max_speed)} does not match number of entities {self.num_entities}"
        else:
            self.max_speed = jnp.concatenate(
                [jnp.full((self.num_agents), -1), jnp.full((self.num_landmarks), 0.0)]
            )

        if "u_noise" in kwargs:
            self.u_noise = kwargs["u_noise"]
            assert (
                len(self.u_noise) == self.num_agents
            ), f"U noise array length {len(self.u_noise)} does not match number of agents {self.num_agents}"
        else:
            self.u_noise = jnp.full((self.num_agents), 0)

        if "c_noise" in kwargs:
            self.c_noise = kwargs["c_noise"]
            assert (
                len(self.c_noise) == self.num_agents
            ), f"C noise array length {len(self.c_noise)} does not match number of agents {self.num_agents}"
        else:
            self.c_noise = jnp.full((self.num_agents), 0)

        if "damping" in kwargs:
            self.damping = kwargs["damping"]
            assert (
                self.damping >= 0
            ), f"Damping must be non-negative, got {self.damping}"
        else:
            self.damping = DAMPING

        if "contact_force" in kwargs:
            self.contact_force = kwargs["contact_force"]
        else:
            self.contact_force = CONTACT_FORCE

        if "contact_margin" in kwargs:
            self.contact_margin = kwargs["contact_margin"]
        else:
            self.contact_margin = CONTACT_MARGIN

    @partial(jax.jit, static_argnums=[0])
    def step_env(self, key: chex.PRNGKey, state: State, actions: dict):
        u, c = self.set_actions(actions)
        if (
            c.shape[1] < self.dim_c
        ):  # This is due to the MPE code carrying around 0s for the communication channels
            c = jnp.concatenate(
                [c, jnp.zeros((self.num_agents, self.dim_c - c.shape[1]))], axis=1
            )

        key, key_w = jax.random.split(key)
        p_pos, p_vel = self._world_step(key_w, state, u)

        key_c = jax.random.split(key, self.num_agents)
        c = self._apply_comm_action(key_c, c, self.c_noise, self.silent)
        done = jnp.full((self.num_agents), state.step >= self.max_steps)

        state = state.replace(
            p_pos=p_pos,
            p_vel=p_vel,
            c=c,
            done=done,
            step=state.step + 1,
        )

        reward = self.rewards(state)

        obs = self.get_obs(state)

        info = {}

        dones = {a: done[i] for i, a in enumerate(self.agents)}
        dones.update({"__all__": jnp.all(done)})

        return obs, state, reward, dones, info

    @partial(jax.jit, static_argnums=[0])
    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        """Initialise with random positions"""

        key_a, key_l = jax.random.split(key)

        p_pos = jnp.concatenate(
            [
                jax.random.uniform(
                    key_a, (self.num_agents, 2), minval=-1.0, maxval=+1.0
                ),
                jax.random.uniform(
                    key_l, (self.num_landmarks, 2), minval=-1.0, maxval=+1.0
                ),
            ]
        )

        state = State(
            p_pos=p_pos,
            p_vel=jnp.zeros((self.num_entities, self.dim_p)),
            c=jnp.zeros((self.num_agents, self.dim_c)),
            done=jnp.full((self.num_agents), False),
            step=0,
        )

        return self.get_obs(state), state

    @partial(jax.jit, static_argnums=[0])
    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        """Return dictionary of agent observations"""

        @partial(jax.vmap, in_axes=[0, None])
        def _observation(aidx: int, state: State) -> jnp.ndarray:
            """Return observation for agent i."""
            landmark_rel_pos = state.p_pos[self.num_agents :] - state.p_pos[aidx]

            return jnp.concatenate(
                [state.p_vel[aidx].flatten(), landmark_rel_pos.flatten()]
            )

        obs = _observation(self.agent_range, state)
        return {a: obs[i] for i, a in enumerate(self.agents)}

    def rewards(self, state: State) -> Dict[str, float]:
        """Assign rewards for all agents"""

        @partial(jax.vmap, in_axes=[0, None])
        def _reward(aidx: int, state: State):
            return -1 * jnp.sum(
                jnp.square(state.p_pos[aidx] - state.p_pos[self.num_agents :])
            )

        r = _reward(self.agent_range, state)
        return {agent: r[i] for i, agent in enumerate(self.agents)}

    def set_actions(self, actions: Dict):
        """Extract u and c actions for all agents from actions Dict."""

        actions = jnp.array([actions[i] for i in self.agents]).reshape(
            (self.num_agents, -1)
        )

        return self.action_decoder(self.agent_range, actions)

    @partial(jax.vmap, in_axes=[None, 0, 0])
    def _decode_continuous_action(
        self, a_idx: int, action: chex.Array
    ) -> Tuple[chex.Array, chex.Array]:
        u = jnp.array([action[1] - action[2], action[3] - action[4]])

        u = u * self.accel[a_idx] * self.moveable[a_idx]
        c = action[5:]
        return u, c

    @partial(jax.vmap, in_axes=[None, 0, 0])
    def _decode_discrete_action(
        self, a_idx: int, action: chex.Array
    ) -> Tuple[chex.Array, chex.Array]:
        u = jnp.zeros((self.dim_p,))
        idx = jax.lax.select(action <= 2, 0, 1)
        u_val = jax.lax.select(action % 2 == 0, 1.0, -1.0) * (action != 0)
        u = u.at[idx].set(u_val)
        u = u * self.accel[a_idx] * self.moveable[a_idx]
        return u, jnp.zeros((self.dim_c,))

    def _world_step(self, key: chex.PRNGKey, state: State, u: chex.Array):
        p_force = jnp.zeros((self.num_agents, 2))

        # apply agent physical controls
        key_noise = jax.random.split(key, self.num_agents)
        p_force = self._apply_action_force(
            key_noise, p_force, u, self.u_noise, self.moveable[: self.num_agents]
        )
        # jax.debug.print('jax p_force post agent {p_force}', p_force=p_force)

        # apply environment forces
        p_force = jnp.concatenate([p_force, jnp.zeros((self.num_landmarks, 2))])
        p_force = self._apply_environment_force(p_force, state)
        # print('p_force post apply env force', p_force)
        # jax.debug.print('jax p_force final: {p_force}', p_force=p_force)

        # integrate physical state
        p_pos, p_vel = self._integrate_state(
            p_force, state.p_pos, state.p_vel, self.mass, self.moveable, self.max_speed
        )

        return p_pos, p_vel

    @partial(jax.vmap, in_axes=[None, 0, 0, 0, 0])
    def _apply_comm_action(
        self, key: chex.PRNGKey, c: chex.Array, c_noise: int, silent: int
    ) -> chex.Array:
        silence = jnp.zeros(c.shape)
        noise = jax.random.normal(key, shape=c.shape) * c_noise
        return jax.lax.select(silent, silence, c + noise)

    # gather agent action forces
    @partial(jax.vmap, in_axes=[None, 0, 0, 0, 0, 0])
    def _apply_action_force(
        self,
        key: chex.PRNGKey,
        p_force: chex.Array,
        u: chex.Array,
        u_noise: int,
        moveable: bool,
    ):
        noise = jax.random.normal(key, shape=u.shape) * u_noise
        return jax.lax.select(moveable, u + noise, p_force)

    def _apply_environment_force(self, p_force_all: chex.Array, state: State):
        """gather physical forces acting on entities"""

        @partial(jax.vmap, in_axes=[0])
        def __env_force_outer(idx: int):
            @partial(jax.vmap, in_axes=[None, 0])
            def __env_force_inner(idx_a: int, idx_b: int):
                l = idx_b <= idx_a
                l_a = jnp.zeros((2, 2))

                collision_force = self._get_collision_force(idx_a, idx_b, state)

                xx = jax.lax.select(l, l_a, collision_force)
                # jax.debug.print('{a} {b} {f}', a=idx_a, b=idx_b, f=xx)
                return xx

            p_force_t = __env_force_inner(idx, self.entity_range)

            p_force_a = jnp.sum(p_force_t[:, 0], axis=0)  # ego force from other agents
            p_force_o = p_force_t[:, 1]
            p_force_o = p_force_o.at[idx].set(p_force_a)

            return p_force_o

        p_forces = __env_force_outer(self.entity_range)
        p_forces = jnp.sum(p_forces, axis=0)

        return p_forces + p_force_all

    @partial(jax.vmap, in_axes=[None, 0, 0, 0, 0, 0, 0])
    def _integrate_state(self, p_force, p_pos, p_vel, mass, moveable, max_speed):
        """integrate physical state"""

        p_pos += p_vel * self.dt
        p_vel = p_vel * (1 - self.damping)

        p_vel += (p_force / mass) * self.dt * moveable

        speed = jnp.sqrt(jnp.square(p_vel[0]) + jnp.square(p_vel[1]))
        over_max = (
            p_vel / jnp.sqrt(jnp.square(p_vel[0]) + jnp.square(p_vel[1])) * max_speed
        )

        p_vel = jax.lax.select((speed > max_speed) & (max_speed >= 0), over_max, p_vel)

        return p_pos, p_vel

    # get collision forces for any contact between two entities BUG
    def _get_collision_force(self, idx_a: int, idx_b: int, state: State):
        dist_min = self.rad[idx_a] + self.rad[idx_b]
        delta_pos = state.p_pos[idx_a] - state.p_pos[idx_b]

        dist = jnp.sqrt(jnp.sum(jnp.square(delta_pos)))

        # softmax penetration
        k = self.contact_margin
        penetration = jnp.logaddexp(0, -(dist - dist_min) / k) * k
        force = self.contact_force * delta_pos / dist * penetration
        force_a = +force * self.moveable[idx_a]
        force_b = -force * self.moveable[idx_b]
        force = jnp.array([force_a, force_b])

        c = (~self.collide[idx_a]) | (~self.collide[idx_b]) | (idx_a == idx_b)
        c_force = jnp.zeros((2, 2))
        return jax.lax.select(c, c_force, force)

    def create_agent_classes(self):
        if hasattr(self, "leader"):
            return {
                "leadadversary": self.leader,
                "adversaries": self.adversaries,
                "agents": self.good_agents,
            }
        elif hasattr(self, "adversaries"):
            return {
                "adversaries": self.adversaries,
                "agents": self.good_agents,
            }
        else:
            return {
                "agents": self.agents,
            }

    def agent_classes(self) -> Dict[str, list]:
        return self.classes

    ### === UTILITIES === ###
    def is_collision(self, a: int, b: int, state: State):
        """check if two entities are colliding"""
        dist_min = self.rad[a] + self.rad[b]
        delta_pos = state.p_pos[a] - state.p_pos[b]
        dist = jnp.sqrt(jnp.sum(jnp.square(delta_pos)))
        return (dist < dist_min) & (self.collide[a] & self.collide[b]) & (a != b)

    @partial(jax.vmap, in_axes=(None, 0))
    def map_bounds_reward(self, x: float):
        """vmap over x, y coodinates"""
        w = x < 0.9
        m = x < 1.0
        mr = (x - 0.9) * 10
        br = jnp.min(jnp.array([jnp.exp(2 * x - 2), 10]))
        return jax.lax.select(m, mr, br) * ~w   


if __name__ == "__main__":
    from jaxmarl.environments.mpe import MPEVisualizer

    num_agents = 3
    key = jax.random.PRNGKey(0)

    env = SimpleMPE(num_agents)

    obs, state = env.reset(key)

    mock_action = jnp.array([[1.0, 1.0, 0.1, 0.1, 0.0]])

    actions = jnp.repeat(mock_action[None], repeats=num_agents, axis=0).squeeze()

    actions = {agent: mock_action for agent in env.agents}
    a = env.agents
    a.reverse()
    print("a", a)
    actions = {agent: mock_action for agent in a}
    print("actions", actions)

    # env.enable_render()

    state_seq = []
    print("state", state)
    print("action spaces", env.action_spaces)

    for _ in range(25):
        state_seq.append(state)
        key, key_act = jax.random.split(key)
        key_act = jax.random.split(key_act, env.num_agents)
        actions = {
            agent: env.action_space(agent).sample(key_act[i])
            for i, agent in enumerate(env.agents)
        }

        obs, state, rew, dones, _ = env.step_env(key, state, actions)

    viz = MPEVisualizer(env, state_seq)
    viz.animate(None, view=True)



import argparse
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np

from jaxmarl.gridworld.maze import Maze #, Actions
from jaxmarl.gridworld.ma_maze import MAMaze
from jaxmarl.environments.overcooked.overcooked import Overcooked
from jaxmarl.environments.overcooked.layouts import layouts


def redraw(state, obs, extras):
    extras['viz'].render(extras['params'], state, highlight=False)

    if extras['obs_viz'] is not None:
        if extras['env'] == "MAMaze" or "Overcooked":
            obs_viz.render_grid(np.asarray(obs['image'][0]), k_rot90=3, agent_dir_idx=[3])
            obs_viz2.render_grid(np.asarray(obs['image'][1]), k_rot90=3, agent_dir_idx=[3])
        else:
            obs_viz.render_grid(np.asarray(obs['image']), k_rot90=3, agent_dir_idx=3)


def reset(key, env, extras):
    key, subkey = jax.random.split(extras['rng'])
    obs, state = extras['jit_reset'](subkey)

    extras['rng'] = key
    extras['obs'] = obs
    extras['state'] = state

    redraw(state, obs, extras)


def step(env, action, extras):
    # TODO: Handle actions better (e.g. choose which agent to control)
    key, subkey = jax.random.split(extras['rng'])

    print("action:", jnp.array([action, action.left]))
    obs, state, reward, done, info = jax.jit(env.step_env)(subkey, extras['state'], jnp.array([action, action]))
    extras['obs'] = obs
    extras['state'] = state
    print(f"reward={reward}, agent_dir={obs['agent_dir']}, agent_inv={state.agent_inv}")
    
    if extras["debug"]:
        layers = [f"player_{i}_loc" for i in range(2)]
        layers.extend([f"player_{i // 4}_orientation_{i % 4}" for i in range(8)])
        layers.extend([
            "pot_loc",
            "counter_loc",
            "onion_disp_loc",
            "tomato_disp_loc",
            "plate_disp_loc",
            "serve_loc",
            "onions_in_pot",
            "tomatoes_in_pot",
            "onions_in_soup",
            "tomatoes_in_soup",
            "soup_cook_time_remaining",
            "soup_done",
            "plates",
            "onions",
            "tomatoes",
            "urgency"
        ])
        print("obs_shape: ", obs["image"].shape)
        print("OBS: \n", obs["image"][1])
        debug_obs = jnp.transpose(obs["image"][1], (2,0,1))
        for i, layer in enumerate(layers):
            print(layer)
            print(debug_obs[i])
    # print(f"agent obs =\n {obs}")

    if done or (jnp.array([action, action]) == Actions.done).any():
        key, subkey = jax.random.split(subkey)
        reset(subkey, env, extras)
    else:
        redraw(state, obs, extras)

    extras['rng'] = key


def key_handler(env, extras, event):
    print('pressed', event.key)

    if event.key == 'escape':
        window.close()
        return

    if event.key == 'backspace':
        extras['jit_reset']((env, extras))
        return

    if event.key == 'left':
        step(env, Actions.left, extras)
        return
    if event.key == 'right':
        step(env, Actions.right, extras)
        return
    if event.key == 'up':
        step(env, Actions.forward, extras)
        return

    # Spacebar
    if event.key == ' ':
        step(env, Actions.toggle, extras)
        return
    if event.key == '[':
        step(env, Actions.pickup, extras)
        return
    if event.key == ']':
        step(env, Actions.drop, extras)
        return

    if event.key == 'enter':
        step(env, Actions.done, extras)
        return

def key_handler_overcooked(env, extras, event):
    print('pressed', event.key)

    if event.key == 'escape':
        window.close()
        return
    if event.key == 'backspace':
        extras['jit_reset']((env, extras))
        return

    if event.key == 'left':
        step(env, Actions.left, extras)
        return
    if event.key == 'right':
        step(env, Actions.right, extras)
        return
    if event.key == 'up':
        # step(env, Actions.forward, extras)
        step(env, Actions.up, extras)
        return
    if event.key == 'down':
        step(env, Actions.down, extras)
        return

    # Spacebar
    if event.key == ' ':
        step(env, Actions.interact, extras)
        return
    if event.key == 'tab':
        step(env, Actions.stay, extras)
        return
    if event.key == 'enter':
        step(env, Actions.done, extras)
        return


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--env",
        type=str,
        help="Environment name",
        default="Overcooked"
    )
    parser.add_argument(
        "--layout",
        type=str,
        help="Overcooked layout",
        default=""
    )
    parser.add_argument(
        "--seed",
        type=int,
        help="random seed to generate the environment with",
        default=0
    )
    parser.add_argument(
        '--render_agent_view',
        default=False,
        help="draw the agent sees (partially observable view)",
        action='store_true'
    )
    parser.add_argument(
        '--height',
        default=13,
        type=int,
        help="height",
    )
    parser.add_argument(
        '--width',
        default=13,
        type=int,
        help="width",
    )
    parser.add_argument(
        '--n_walls',
        default=50,
        type=int,
        help="Number of walls",
    )
    parser.add_argument(
        '--agent_view_size',
        default=5,
        type=int,
        help="Number of walls",
    )
    parser.add_argument(
        '--debug',
        default=False,
        help="Debug mode",
        action='store_true'
    )
    args = parser.parse_args()

    if args.env == "Maze":
        env = Maze(
            height=13,
            width=13,
            n_walls=25,
            see_agent=True,
        )
        from jaxmarl.gridworld.grid_viz import GridVisualizer as Visualizer
        from jaxmarl.gridworld.maze import Actions

    elif args.env == "MAMaze":
        env = MAMaze(
            height=13,
            width=13,
            n_walls=25,
            see_agent=True,
            n_agents=2
        )
        from jaxmarl.gridworld.grid_viz import GridVisualizer as Visualizer
        from jaxmarl.gridworld.maze import Actions

    elif args.env == "Overcooked":
        if len(args.layout) > 0:
            layout = layouts[args.layout]
            env = Overcooked(
                height=layout["height"],
                width=layout["width"],
                n_walls=1,
                see_agent=True,
                n_agents=2,
                fixed_layout=True,
                layout=layout
            )
        else:
            env = Overcooked(
                height=13,
                width=13,
                n_walls=25,
                see_agent=True,
                n_agents=2
            )
        from jaxmarl.viz.overcooked_visualizer import OvercookedVisualizer as Visualizer
        from jaxmarl.environments.overcooked.overcooked import Actions

    params = env.params

    viz = Visualizer()
    obs_viz = None
    obs_viz2 = None
    if args.render_agent_view:
        obs_viz = Visualizer()
        if args.env == "MAMaze" or "Overcooked":
            obs_viz2 = Visualizer()

    with jax.disable_jit(True):
        jit_reset = jax.jit(env.reset_env, static_argnums=(1,))
        # jit_reset = env.reset_env
        key = jax.random.PRNGKey(args.seed)
        key, subkey = jax.random.split(key)
        o0, s0 = jit_reset(subkey)
        viz.render(params, s0, highlight=False)
        if obs_viz is not None:
            if args.env == "MAMaze" or args.env == "Overcooked":
                obs_viz.render_grid(np.asarray(o0['image'][0]), k_rot90=3, agent_dir_idx=[3])
                obs_viz2.render_grid(np.asarray(o0['image'][1]), k_rot90=3, agent_dir_idx=[3])
            else:
                obs_viz.render_grid(np.asarray(o0['image']), k_rot90=3, agent_dir_idx=3)

        key, subkey = jax.random.split(key)
        extras = {
            'rng': subkey,
            'state': s0,
            'obs': o0,
            'params': params,
            'viz': viz,
            'obs_viz': obs_viz,
            'obs_viz2': obs_viz2,
            'jit_reset': jit_reset,
            'env': args.env,
            'debug': args.debug
        }

        if args.env == "Overcooked":
            viz.window.reg_key_handler(partial(key_handler_overcooked, env, extras))
            viz.show(block=True)
        else:
            viz.window.reg_key_handler(partial(key_handler, env, extras))
            viz.show(block=True)
